import numpy as np
import warnings


class ConjugateGradients:
    """
     `[Source] <https://github.com/ESFIEP/EarlyStopping/edit/main/src/EarlyStopping/conjugate_gradients.py>`_ A class to perform estimation using the conjugate gradients algorithm for the normal equation.

    **Description**

    Consider the *linear model*

    .. math::
        Y = Af + \\delta Z,

    where :math:`Z` is a :math:`n`-dimensional standard normal distribution.
    The conjugate gradient estimate :math:`\\hat{f}^{(m)}` at the integer iteration index :math:`m` is iteratively calculated by the *conjugate gradients for the normal equation* algorithm with initial value :math:`\\hat{f}_0`,
    see `Björck (1996, Algorithm 7.4.1) <https://doi.org/10.1137/1.9781611971484>`_.

    **Parameters**

    *design*: ``ndarray``. nxp design matrix of the linear model. ( :math:`A \\in \\mathbb{R}^{n \\times p}` )

    *response*: ``ndarray``. n-dim vector of the observed data in the linear model. ( :math:`Y \in \mathbb{R}^{n}` )

    *initial_value*: ``array, default = None``. Determines the zeroth step of the iterative procedure. Default is zero. ( :math:`\\hat{f}_0` )

    *true_signal*: ``ndarray, default = None``. p-dim vector for simulation purposes only. For simulated data the true signal can be included to compute additional quantities alongside the iterative procedure. ( :math:`f \\in \\mathbb{R}^{p}` )

    *true_noise_level*: ``float, default = None``. For simulation purposes only. Corresponds to the standard deviation of normally distributed noise contributing to the response variable. ( :math:`\\delta \\geq 0` )

    *computation_threshold*: ``float, default = 10 ** (-8)``. Threshold used to terminate the conjugate gradients algorithm.

    **Attributes**

    *sample_size*: ``int``. Sample size of the linear model. ( :math:`n \\in \\mathbb{N}` )

    *parameter_size*: ``int``. Parameter size of the linear model. ( :math:`p \\in \\mathbb{N}` )

    *iteration*: ``int``. Current conjugate gradient iteration of the algorithm. ( :math:`m \\in \\mathbb{N}` )

    *conjugate_gradient_estimate*: ``ndarray``. Conjugate gradient estimate at the current iteration for the data given in design and response. ( :math:`\\hat{f}^{(m)}` )

    *conjugate_gradient_estimate_list*: ``list``. List containing the conjugate gradient estimates at integer iteration indices up to the current conjugate gradient iteration.

    *residuals*: ``ndarray``. Lists the sequence of the squared residuals between the observed data and the conjugate gradient estimator.

    *strong_empirical_risk*: ``ndarray``. Only exists if true_signal was given. Lists the values of the strong empirical error between the conjugate gradient estimator and the true signal up to the current conjugate gradient iteration.

    *weak_empirical_risk*: ``ndarray``. Only exists if true_signal was given. Lists the values of the weak empirical error between the conjugate gradient estimator and the true signal up to the current conjugate gradient iteration.

    **Methods**

    +---------------------------------------------------------------------------------------+---------------------------------------------------------------------------------------------+
    | iterate(``number_of_iterations = 1``)                                                 | Performs a specified number of iterations of the conjugate gradients algorithm.             |
    +---------------------------------------------------------------------------------------+---------------------------------------------------------------------------------------------+
    | get_estimate(``iteration``)                                                           | Returns the conjugate gradient estimator at iteration.                                      |
    +---------------------------------------------------------------------------------------+---------------------------------------------------------------------------------------------+
    | get_discrepancy_stop(``critical_value``, ``max_iteration``, ``interpolation = False``)| Returns the early stopping index according to the discrepancy principle with emergency stop.|
    +---------------------------------------------------------------------------------------+---------------------------------------------------------------------------------------------+
    | get_residual(``iteration``)                                                           | Returns the squared residual at iteration.                                                  |
    +---------------------------------------------------------------------------------------+---------------------------------------------------------------------------------------------+
    | get_strong_empirical_risk(``iteration``)                                              | Returns the strong empirical error at iteration.                                            |
    +---------------------------------------------------------------------------------------+---------------------------------------------------------------------------------------------+
    | get_weak_empirical_risk(``iteration``)                                                | Returns the weak empirical error at iteration.                                              |
    +---------------------------------------------------------------------------------------+---------------------------------------------------------------------------------------------+
    | get_strong_empirical_oracle(``max_iteration``, ``interpolation = False``)             | Returns the strong empirical oracle up to max_iteration.                                    |
    +---------------------------------------------------------------------------------------+---------------------------------------------------------------------------------------------+
    | get_weak_empirical_oracle(``max_iteration``, ``interpolation = False``)               | Returns the weak empirical oracle up to max_iteration.                                      |
    +---------------------------------------------------------------------------------------+---------------------------------------------------------------------------------------------+
    """

    def __init__(
        self,
        design,
        response,
        initial_value=None,
        true_signal=None,
        true_noise_level=None,
        computation_threshold=10 ** (-8),
    ):
        self.design = design
        self.response = response
        self.initial_value = initial_value
        self.true_signal = true_signal
        self.true_noise_level = true_noise_level

        self.computation_threshold = computation_threshold

        # Parameters of the model
        self.sample_size = np.shape(design)[0]
        self.parameter_size = np.shape(design)[1]

        # Determine starting value for the procedure
        if initial_value is None:
            warnings.warn("No initial_value is given, using zero by default.", category=UserWarning)
            self.initial_value = np.zeros(self.parameter_size)
        else:
            if initial_value.size != self.parameter_size:
                raise ValueError("The dimension of the initial_value should match paramter size!")
            self.initial_value = initial_value

        # Estimation quantities
        self.iteration = 0
        self.conjugate_gradient_estimate = self.initial_value

        # Collect coefficients at integer iteration indices
        self.conjugate_gradient_estimate_list = [self.initial_value]

        # Residual quantities
        self.__residual_vector = self.response - self.design @ self.initial_value
        self.residuals = np.array([np.sum(self.__residual_vector**2)])
        self.__transformed_residuals = np.array([np.sum((np.transpose(self.design) @ self.__residual_vector) ** 2)])

        # Initialize theoretical quantities
        if self.true_signal is not None:
            self.__transformed_true_signal = self.design @ self.true_signal
            self.strong_empirical_risk = np.array([np.sum((self.initial_value - self.true_signal) ** 2)])
            self.weak_empirical_risk = np.array(
                [np.sum((self.design @ self.initial_value - self.__transformed_true_signal) ** 2)]
            )
            self.__strong_estimator_distances = np.array([0])
            self.__weak_estimator_distances = np.array([0])

        # Starting values for the algorithm
        self.__transposed_design = np.transpose(design)
        self.__transformed_residual_vector = self.__transposed_design @ self.__residual_vector
        self.__search_direction = self.__transformed_residual_vector

    def iterate(self, number_of_iterations=1):
        """Performs number_of_iterations iterations of the conjugate gradients algorithm

        **Parameters**

        *number_of_iterations*: ``int``. The number of iterations to perform.
        """
        for _ in range(number_of_iterations):
            if self.__transformed_residuals[self.iteration] <= self.computation_threshold:
                warnings.warn(
                    f"Algorithm terminates at iteration {self.iteration}: norm of transformed residual vector ({self.__transformed_residuals[self.iteration]}) <= computation_threshold ({self.computation_threshold}).",
                    category=UserWarning,
                )
                break
            self.__conjugate_gradients_one_iteration()

    def get_estimate(self, iteration):
        """Returns the conjugate gradient estimate at a possibly noninteger iteration

        **Parameters**

        *iteration*: ``int or float``. The (possibly noninteger) iteration at which the conjugate gradient estimate is requested.

        **Returns**

        *conjugate_gradient_estimate*: ``ndarray``. The conjugate gradient estimate at iteration.
        """
        if iteration is None:
            raise ValueError("iteration is None. Potentially from querying the estimate at an oracle or stopping time that was not found until max_iteration.")

        iteration_ceil = np.ceil(iteration).astype("int")

        if iteration_ceil > self.iteration:
            self.iterate(iteration_ceil - self.iteration)

        if iteration_ceil > self.iteration:
            raise ValueError(
                "Algorithm terminated due to computation_threshold before (ceiling of) requested iteration."
            )

        if iteration % 1 == 0:
            conjugate_gradient_estimate = self.conjugate_gradient_estimate_list[int(iteration)]
        else:
            iteration_floor = np.floor(iteration).astype("int")
            alpha = iteration - iteration_floor
            conjugate_gradient_estimate = (1 - alpha) * self.conjugate_gradient_estimate_list[
                iteration_floor
            ] + alpha * self.conjugate_gradient_estimate_list[iteration_ceil]

        return conjugate_gradient_estimate

    def get_discrepancy_stop(self, critical_value, max_iteration, interpolation=False):
        """Returns early stopping index based on discrepancy principle up to max_iteration

        **Parameters**

        *critical_value*: ``float``. The critical value for the discrepancy principle. The algorithm stops when
        :math:`\\Vert Y - A \\hat{f}^{(m)} \\Vert^{2} \\leq \\kappa^{2},`
        where :math:`\\kappa` is the critical value.

        *max_iteration*: ``int``. The maximum number of total iterations to be considered.

        *interpolation*: ``boolean, default = False``. If interpolation is set to ``True``, the early stopping index can be noninteger valued.

        **Returns**

        *early_stopping_index*: ``int or float``. The first iteration at which the discrepancy principle is satisfied or
        :math:`\\Vert A^{\\top}(Y - A \\hat{f}^{(m)}) \\Vert^{2} \\leq C,` where :math:`C` is the computation_threshold
        for the emergency stop.
        (``None`` is returned if the stopping index is not found.)
        """

        # Discrepancy stop on existing estimators
        if np.any(self.residuals <= critical_value):
            iteration = np.argmax(self.residuals <= critical_value)
            if interpolation is True and iteration > 0:
                alpha = 1 - np.sqrt(
                    1
                    - (self.residuals[iteration - 1] - critical_value)
                    / (self.residuals[iteration - 1] - self.residuals[iteration])
                )
                early_stopping_index = iteration - 1 + alpha
            else:
                early_stopping_index = iteration
            return early_stopping_index
        # Further iteration if necessary and possible
        elif self.__transformed_residuals[self.iteration] > self.computation_threshold:
            while (
                self.residuals[self.iteration] > critical_value
                and self.__transformed_residuals[self.iteration] > self.computation_threshold
                and self.iteration < max_iteration
            ):
                self.__conjugate_gradients_one_iteration()

        # Emergency stop
        if self.__transformed_residuals[self.iteration] <= self.computation_threshold:
            early_stopping_index = self.iteration
            return early_stopping_index

        # Discrepancy stop
        if self.residuals[self.iteration] <= critical_value:
            if interpolation is True and self.iteration > 0:
                alpha = 1 - np.sqrt(
                    1
                    - (self.residuals[self.iteration - 1] - critical_value)
                    / (self.residuals[self.iteration - 1] - self.residuals[self.iteration])
                )
                early_stopping_index = self.iteration - 1 + alpha
            else:
                early_stopping_index = self.iteration
            return early_stopping_index
        else:
            warnings.warn("Early stopping index not found up to max_iteration. Returning None.", category=UserWarning)
            return None

    def get_residual(self, iteration):
        """Returns the squared residual at a possibly noninteger iteration index

        **Parameters**

        *iteration*: ``int or float``. Iteration index where the squared residual should be calculated.

        **Returns**

        *residual*: ``float``. The squared residual at the requested iteration index.
        """

        iteration_ceil = np.ceil(iteration).astype("int")

        if iteration_ceil > self.iteration:
            self.iterate(iteration_ceil - self.iteration)

        if iteration_ceil > self.iteration:
            raise ValueError(
                "Algorithm terminated due to computation_threshold before (ceiling of) requested iteration."
            )

        if iteration % 1 == 0:
            residual = self.residuals[int(iteration)]
        else:
            iteration_floor = np.floor(iteration).astype("int")
            alpha = iteration - iteration_floor
            residual = (1 - alpha) ** 2 * self.residuals[iteration_floor] + (1 - (1 - alpha) ** 2) * self.residuals[
                iteration_ceil
            ]
        return residual

    def get_strong_empirical_risk(self, iteration):
        """Returns the strong empirical error at a possibly noninteger iteration index

        **Parameters**

        *iteration*: ``int or float``. Iteration index where the error should be calculated.

        **Returns**

        *strong_empirical_risk*: ``float``. The strong empirical error at the requested iteration index.
        """

        if self.true_signal is None:
            raise ValueError("No true signal given.")

        conjugate_gradient_estimate = self.get_estimate(iteration)

        strong_empirical_risk = np.sum((conjugate_gradient_estimate - self.true_signal) ** 2)
        return strong_empirical_risk

    def get_weak_empirical_risk(self, iteration):
        """Returns the weak empirical error at a possibly noninteger iteration index

        **Parameters**

        *iteration*: ``int or float``. Iteration index where the error should be calculated.

        **Returns**

        *weak_empirical_risk*: ``float``. The weak empirical error at the requested iteration index.
        """

        if self.true_signal is None:
            raise ValueError("No true signal given.")

        conjugate_gradient_estimate = self.get_estimate(iteration)

        weak_empirical_risk = np.sum((self.design @ conjugate_gradient_estimate - self.__transformed_true_signal) ** 2)
        return weak_empirical_risk

    def get_strong_empirical_oracle(self, max_iteration, interpolation=False):
        """Returns the strong empirical oracle up to max_iteration

        **Parameters**

        *max_iteration*: ``int``. The maximum number of total iterations to be considered.

        *interpolation*: ``boolean, default = False``. If interpolation is set to ``True``, the strong empirical oracle can be noninteger valued.

        **Returns**

        *strong_empirical_oracle*: ``int or float``. The iteration index at which the strong empirical error is minimal along the iteration path up to max_iteration. If not unique, the smallest one is returned.
        """

        if self.true_signal is None:
            raise ValueError("No true signal given.")

        if max_iteration > self.iteration:
            self.iterate(max_iteration - self.iteration)

        if max_iteration > self.iteration:
            warnings.warn(
                "Algorithm terminated due to computation_threshold before max_iteration. max_iteration is set to terminal iteration index.",
                category=UserWarning,
            )
            max_iteration = self.iteration

        if interpolation is True:
            optimal_index_list = []
            empirical_risk_at_optimal_index_list = []
            empirical_risk = self.strong_empirical_risk
            estimator_distances = self.__strong_estimator_distances
            for index in np.arange(max_iteration):
                if np.sqrt(estimator_distances[index + 1]) <= self.computation_threshold:
                    alpha = 0
                else:
                    alpha = (empirical_risk[index] - empirical_risk[index + 1] + estimator_distances[index + 1]) / (
                        2 * estimator_distances[index + 1]
                    )
                if alpha < 0:
                    optimal_index_candidate = index
                elif alpha > 1:
                    optimal_index_candidate = index + 1
                else:
                    optimal_index_candidate = index + alpha
                optimal_index_list = np.append(optimal_index_list, optimal_index_candidate)
                empirical_risk_at_optimal_index_list = np.append(
                    empirical_risk_at_optimal_index_list,
                    self.get_strong_empirical_risk(optimal_index_candidate),
                )
            strong_empirical_oracle = optimal_index_list[np.argmin(empirical_risk_at_optimal_index_list)]
        else:
            strong_empirical_oracle = np.argmin(self.strong_empirical_risk)

        return strong_empirical_oracle

    def get_weak_empirical_oracle(self, max_iteration, interpolation=False):
        """Returns the weak empirical oracle up to max_iteration

        **Parameters**

        *max_iteration*: ``int``. The maximum number of total iterations to be considered.

        *interpolation*: ``boolean, default = False``. If interpolation is set to ``True``, the weak empirical oracle can be noninteger valued.

        **Returns**

        *weak_empirical_oracle*: ``int or float``. The iteration index at which the weak empirical error is minimal along the iteration path up to max_iteration. If not unique, the smallest one is returned.
        """

        if self.true_signal is None:
            raise ValueError("No true signal given.")

        if max_iteration > self.iteration:
            self.iterate(max_iteration - self.iteration)

        if max_iteration > self.iteration:
            warnings.warn(
                "Algorithm terminated due to computation_threshold before max_iteration. max_iteration is set to terminal iteration index.",
                category=UserWarning,
            )
            max_iteration = self.iteration

        if interpolation is True:
            optimal_index_list = []
            empirical_risk_at_optimal_index_list = []
            empirical_risk = self.weak_empirical_risk
            estimator_distances = self.__weak_estimator_distances
            for index in np.arange(max_iteration):
                if np.sqrt(estimator_distances[index + 1]) <= self.computation_threshold:
                    alpha = 0
                else:
                    alpha = (empirical_risk[index] - empirical_risk[index + 1] + estimator_distances[index + 1]) / (
                        2 * estimator_distances[index + 1]
                    )
                if alpha < 0:
                    optimal_index_candidate = index
                elif alpha > 1:
                    optimal_index_candidate = index + 1
                else:
                    optimal_index_candidate = index + alpha
                optimal_index_list = np.append(optimal_index_list, optimal_index_candidate)
                empirical_risk_at_optimal_index_list = np.append(
                    empirical_risk_at_optimal_index_list,
                    self.get_weak_empirical_risk(optimal_index_candidate),
                )
            weak_empirical_oracle = optimal_index_list[np.argmin(empirical_risk_at_optimal_index_list)]
        else:
            weak_empirical_oracle = np.argmin(self.weak_empirical_risk)
        return weak_empirical_oracle

    def __update_strong_empirical_risk(self):
        """Update the strong empirical error"""
        strong_empirical_risk = np.sum((self.conjugate_gradient_estimate - self.true_signal) ** 2)
        self.strong_empirical_risk = np.append(self.strong_empirical_risk, strong_empirical_risk)

    def __update_weak_empirical_risk(self):
        """Update the weak empirical error"""
        weak_empirical_risk = np.sum(
            (self.design @ self.conjugate_gradient_estimate - self.__transformed_true_signal) ** 2
        )
        self.weak_empirical_risk = np.append(self.weak_empirical_risk, weak_empirical_risk)

    def __conjugate_gradients_one_iteration(self):
        """Performs one iteration of the conjugate gradients algorithm"""
        old_transformed_residual_vector = self.__transformed_residual_vector
        squared_norm_old_transformed_residual_vector = np.sum(old_transformed_residual_vector**2)
        transformed_search_direction = self.design @ self.__search_direction
        learning_rate = squared_norm_old_transformed_residual_vector / np.sum(transformed_search_direction**2)
        conjugate_gradient_estimates_distance = learning_rate * self.__search_direction
        transformed_conjugate_gradient_estimates_distance = learning_rate * transformed_search_direction
        self.conjugate_gradient_estimate = self.conjugate_gradient_estimate + conjugate_gradient_estimates_distance
        self.conjugate_gradient_estimate_list.append(self.conjugate_gradient_estimate)
        self.__residual_vector = self.__residual_vector - transformed_conjugate_gradient_estimates_distance
        self.__transformed_residual_vector = self.__transposed_design @ self.__residual_vector
        transformed_residual_ratio = (
            np.sum(self.__transformed_residual_vector**2) / squared_norm_old_transformed_residual_vector
        )
        self.__search_direction = (
            self.__transformed_residual_vector + transformed_residual_ratio * self.__search_direction
        )
        self.residuals = np.append(self.residuals, np.sum(self.__residual_vector**2))
        self.__transformed_residuals = np.append(
            self.__transformed_residuals, np.sum(self.__transformed_residual_vector**2)
        )

        self.iteration = self.iteration + 1

        if self.true_signal is not None:
            self.__update_strong_empirical_risk()
            self.__update_weak_empirical_risk()
            self.__strong_estimator_distances = np.append(
                self.__strong_estimator_distances,
                np.sum(conjugate_gradient_estimates_distance**2),
            )
            self.__weak_estimator_distances = np.append(
                self.__weak_estimator_distances,
                np.sum(transformed_conjugate_gradient_estimates_distance**2),
            )



import numpy as np
import scipy
from scipy import sparse
from scipy.sparse.linalg import svds
from scipy.sparse import dia_matrix
import warnings

class TruncatedSVD:
    """
    `[Source] <https://github.com/ESFIEP/EarlyStopping/edit/main/src/EarlyStopping/truncated_svd.py>`_ 
    A class to perform estimation using truncated SVD estimation.

    **Parameters**

    *design*: ``array``. Design matrix of the linear model. ( :math:`A \in \mathbb{R}^{D \\times p}` ).

    *response*: ``array``. n-dim vector of the observed data in the linear model. ( :math:`Y \in \mathbb{R}^{n}` ).

    *true_signal*: ``array, default = None``. p-dim vector of the true signal. For simulation purposes only.
    For simulated data, the true signal can be included to compute theoretical quantities such as
    the bias and the risk alongside the iterative procedure. ( :math:`f \in \mathbb{R}^{p}` ).

    *true_noise_level*: ``float, default = None`` For simulation purposes only. Corresponds to the
    standard deviation of normally distributed noise contributing to the response variable. Allows
    the analytic computation of the strong and weak variance. ( :math:`\delta \geq 0` ).

    *diagonal*: ``bool, default = False`` The user may set this to true if the design matrix is
    diagonal with strictly positive singular values to avoid unnecessary computation in the diagonal
    sequence space model.
    # 2024-10-14, Bernhard: Checked docu of the parameters. 

    **Attributes**

    *iteration*: ``int``. Current iteration of the algorithm ( :math:`m \in \mathbb{N}` )

    *sample_size*: ``int``. Sample size of the linear model ( :math:`D \in \mathbb{N}` )

    *parameter_size*: ``int``. Parameter size of the linear model ( :math:`p \in \mathbb{N}` )

    *residuals*: ``array``. Lists the sequence of the squared residuals between the observed data
    and the estimator.

    *weak_bias2*: ``array``. Only exists if true_signal was given. Lists the values of the weak
    squared bias up to the current iteration.

    *weak_variance*: ``array``. Only exists if true_signal was given. Lists the values of the weak
    variance up to the current iteration.

    *weak_risk*: ``array``. Only exists if true_signal was given. Lists the values of the weak
    mean squared error up to the current iteration.

    *strong_bias2*: ``array``. Only exists if true_signal was given. Lists the values of the strong
    squared bias up to the current iteration.

    *strong_variance*: ``array``. Only exists if true_signal was given. Lists the values of the
    strong variance up to the current iteration.

    *strong_risk*: ``array``. Only exists if true_signal was given. Lists the values of the strong
    mean squared error up to the current iteration.

    **Methods**

    +---------------------------------------------------------+--------------------------------------------------------------------------+
    | iterate(``number_of_iterations=1``)                     | Performs a specified number of iterations of the Landweber algorithm.    |
    +---------------------------------------------------------+--------------------------------------------------------------------------+
    | get_estimate(``iteration``)                             | Returns the truncated SVD estimator at iteration.                        |
    +---------------------------------------------------------+--------------------------------------------------------------------------+
    | get_discrepancy_stop(``critical_value, max_iteration``) | Returns the early stopping index according to the discrepancy principle. |
    +---------------------------------------------------------+--------------------------------------------------------------------------+
    | get_aic(``max_iteration, K=2``)                         | Returns the iteration chosen by the Akaike information criterion.        |
    +---------------------------------------------------------+--------------------------------------------------------------------------+
    | get_weak_balanced_oracle(``max_iteration``)             | Returns the weak balanced oracle if found up to max_iteration.           |
    +---------------------------------------------------------+--------------------------------------------------------------------------+
    | get_strong_balanced_oracle(``max_iteration``)           | Returns the strong balanced oracle if found up to max_iteration.         |
    +---------------------------------------------------------+--------------------------------------------------------------------------+

    """
    def __init__(self,
        design,
        response,
        true_signal = None,
        true_noise_level = None,
        diagonal = False
    ):

        self.design = design
        self.response = response
        self.true_signal = true_signal
        self.true_noise_level = true_noise_level
        self.diagonal = diagonal

        # Parameters of the model
        self.sample_size = np.shape(self.design)[0]
        self.parameter_size = np.shape(self.design)[1]

        self.iteration = 0

        # Quantities in terms of the SVD
        self.diagonal_design      = np.array([])
        self.diagonal_true_signal = np.array([])
        self.diagonal_response    = np.array([])
        self.diagonal_estimate    = np.array([])

        self.reduced_design = design
        self.eigenvector_matrix = np.empty((self.parameter_size, 0))

        self.residuals = np.array([np.sum(self.response**2)])
        self.truncated_svd_estimate_list = [np.zeros(self.parameter_size)]

        # Check for sparsity
        if (self.diagonal == True) and (not isinstance(self.design, dia_matrix)):
            raise TypeError("The diagonal design matrix is not of dia_sparse type. Please refine!")
        elif (self.diagonal == True):
            self.diagonal_design = self.design.diagonal()

        # Initialize theoretical quantities
        if self.true_signal is not None:
            self.weak_bias2    = np.array([np.sum((self.design @ self.true_signal)**2)])
            self.weak_variance = np.array([0])
            self.weak_risk      = np.array([np.sum((self.design @ self.true_signal)**2)])

            self.strong_bias2    = np.array([np.sum(self.true_signal**2)])
            self.strong_variance = np.array([0])
            self.strong_risk      = np.array([np.sum(self.true_signal**2)])

    def iterate(self, number_of_iterations):
        """Performs number_of_iterations iterations of the algorithm.

        **Parameters**

        *number_of_iterations*: ``int``. The number of iterations to perform.
        """
        if not self.diagonal: 
            for _ in range(number_of_iterations):
                self.__truncated_SVD_one_iteration()
        else:
            for _ in range(number_of_iterations):
                self.__truncated_SVD_one_iteration_diagonal()

    def get_estimate(self, iteration):
        """Returns the truncated SVD estimate at iteration.

        **Parameters**

        *iteration*: ``int``. The iteration at which the estimate is requested.

        **Returns**

        *truncated_svd_estimate*: ``ndarray``. The truncated svd estimate at iteration.
        """
        if iteration is None:
            raise ValueError("iteration is None. Potentially from querying the estimate at an oracle or stopping time that was not found until max_iteration.")

        if iteration > self.iteration:
            self.iterate(iteration - self.iteration)

        truncated_svd_estimate = self.truncated_svd_estimate_list[iteration]
        return truncated_svd_estimate

    def get_discrepancy_stop(self, critical_value, max_iteration):
        """Returns early stopping index based on discrepancy principle up to max_iteration.

        **Parameters**

        *critical_value*: ``float``. The critical value for the discrepancy principle. The algorithm
        stops when :math: `\\Vert Y - A \hat{f}^{(m)} \\Vert^{2} \leq \\kappa^{2},` where
        :math: `\\kappa` is the critical value.

        *max_iteration*: ``int``. The maximum number of total iterations to be considered.

        **Returns**

        *early_stopping_index*: ``int``. The first iteration at which the discrepancy principle is
        satisfied. (None is returned if the stopping index is not found.)
        """
        if self.residuals[self.iteration] <= critical_value:
            # argmax takes the first instance of True in the true-false array
            early_stopping_index = np.argmax(self.residuals <= critical_value)
            return int(early_stopping_index)

        if self.residuals[self.iteration] > critical_value:
            while self.residuals[self.iteration] > critical_value and self.iteration <= max_iteration:
                self.iterate(1)

        if self.residuals[self.iteration] <= critical_value:
            early_stopping_index = self.iteration
            return early_stopping_index
        else:
            warnings.warn("Early stopping index not found up to max_iteration. Returning None.", category=UserWarning)
            return None

    def get_aic(self, max_iteration, K = 2):
        """Returns the iteration chosen by the Akaike information criterion computed up max_iteration.

        **Parameters**

        *K*: ``float``. Constant in the definition of the AIC.

        *max_iteration*: ``int``. The maximum number of total iterations to be considered.

        **Returns**

        *aic_index*: ``int``. Minimiser of the AIC criterion.
        """
        if max_iteration > self.iteration:
            self.iterate(max_iteration - self.iteration)

        if self.diagonal:
            # It's allowed to overwrite this because it is not used if self.diagonal == True
            self.diagonal_response = self.response

        data_fit_term = - np.cumsum(self.diagonal_response[:max_iteration]**2 / \
                                    self.diagonal_design[:max_iteration]**2)
        penalty_term  = K * self.true_noise_level**2 * \
                            np.cumsum(self.diagonal_design[:max_iteration]**(-2))
        aic           = data_fit_term + penalty_term 
        aic_index     = np.argmin(aic)

        return [aic_index, aic]

    def get_weak_balanced_oracle(self, max_iteration):
        """Returns weak balanced oracle if found up to max_iteration.

        **Parameters**

        *max_iteration*: ``int``. The maximum number of total iterations to be considered.

        **Returns**

        *weak_balanced_oracle*: ``int``. The first iteration at which the weak bias is smaller than
        the weak variance.
        """
        if self.weak_bias2[self.iteration] <= self.weak_variance[self.iteration]:
            # argmax takes the first instance of True in the true-false array
            weak_balanced_oracle = np.argmax(self.weak_bias2 <= self.weak_variance)
            return int(weak_balanced_oracle)

        if self.weak_bias2[self.iteration] > self.weak_variance[self.iteration]:
            while (self.weak_bias2[self.iteration] > self.weak_variance[self.iteration] and
                   self.iteration <= max_iteration
                  ):
                self.iterate(1)

        if self.weak_bias2[self.iteration] <= self.weak_variance[self.iteration]:
            weak_balanced_oracle = self.iteration
            return weak_balanced_oracle
        else:
            warnings.warn("Weakly balanced oracle not found up to max_iteration. Returning None.",
                          category=UserWarning)
            return None

    def get_strong_balanced_oracle(self, max_iteration):
        """Returns strong balanced oracle if found up to max_iteration.

        **Parameters**

        *max_iteration*: ``int``. The maximum number of total iterations to be considered.

        **Returns**

        *strong_balanced_oracle*: ``int``. The first iteration at which the strong bias is smaller
        than the strong variance.
        """
        if self.strong_bias2[self.iteration] <= self.strong_variance[self.iteration]:
            # argmax takes the first instance of True in the true-false array
            strong_balanced_oracle = np.argmax(self.strong_bias2 <= self.strong_variance)
            return int(strong_balanced_oracle)

        if self.strong_bias2[self.iteration] > self.strong_variance[self.iteration]:
            while (self.strong_bias2[self.iteration] > self.strong_variance[self.iteration] and
                   self.iteration <= max_iteration
                  ):
                self.iterate(1)

        if self.strong_bias2[self.iteration] <= self.strong_variance[self.iteration]:
            strong_balanced_oracle = self.iteration
            return strong_balanced_oracle
        else:
            warnings.warn("Weakly balanced oracle not found up to max_iteration. Returning None.",
                          category=UserWarning)
            return None

    def __truncated_SVD_one_iteration(self):
        # Get next singular triplet
        u, s, vh = svds(self.reduced_design, k=1)

        # Get diagonal sequence model quantities
        self.diagonal_design    = np.append(self.diagonal_design, s)
        self.diagonal_response  = np.append(self.diagonal_response, u.transpose() @ self.response)
        self.eigenvector_matrix = np.append(self.eigenvector_matrix, vh.transpose(), axis=1)
        self.diagonal_estimate  = np.append(self.diagonal_estimate,
                                            self.diagonal_response[self.iteration] / s)

        # Update full model quantities
        new_truncated_svd_estimate = self.truncated_svd_estimate_list[self.iteration] + \
                                     self.diagonal_estimate[self.iteration] * vh.flatten()
        self.truncated_svd_estimate_list.append(new_truncated_svd_estimate) 

        new_residual   = self.residuals[self.iteration] - (u.transpose() @ self.response)**2
        self.residuals = np.append(self.residuals, new_residual)

        # Reduce design by one eigen triplet 
        self.reduced_design = self.reduced_design - s * u @ vh

        # Updating theoretical quantities
        if self.true_signal is not None:
            self.diagonal_true_signal = np.append(self.diagonal_true_signal, u.transpose() @ self.true_signal)

            new_weak_bias2  = self.weak_bias2[self.iteration] - s**2 * self.diagonal_true_signal[self.iteration]**2
            self.weak_bias2 = np.append(self.weak_bias2, new_weak_bias2)

            new_weak_variance  = self.weak_variance[self.iteration] + self.true_noise_level**2
            self.weak_variance = np.append(self.weak_variance, new_weak_variance)

            new_weak_risk = new_weak_bias2 + new_weak_variance
            self.weak_risk = np.append(self.weak_risk, new_weak_risk)

            new_strong_bias2  = self.strong_bias2[self.iteration] - self.diagonal_true_signal[self.iteration]**2
            self.strong_bias2 = np.append(self.strong_bias2, new_strong_bias2)

            new_strong_variance = self.strong_variance[self.iteration] + self.true_noise_level**2 / s**2
            self.strong_variance = np.append(self.strong_variance, new_strong_variance)

            new_strong_risk = new_strong_bias2 + new_strong_variance
            self.strong_risk = np.append(self.strong_risk, new_strong_risk)
        
        self.iteration += 1

    def __truncated_SVD_one_iteration_diagonal(self):
        s = self.diagonal_design[self.iteration]

        # Estimator update
        standard_basis_vector                 = np.zeros(self.parameter_size)
        standard_basis_vector[self.iteration] = 1.0
        new_truncated_svd_estimate            = self.truncated_svd_estimate_list[self.iteration] + \
                                                self.response[self.iteration] / s * standard_basis_vector
        self.truncated_svd_estimate_list.append(new_truncated_svd_estimate) 

        # Residual update
        new_residual   = self.residuals[self.iteration] - self.response[self.iteration]**2
        self.residuals = np.append(self.residuals, new_residual)

        # Updating theoretical quantities
        if self.true_signal is not None:
            new_weak_bias2  = self.weak_bias2[self.iteration] - s**2 * self.true_signal[self.iteration]**2
            self.weak_bias2 = np.append(self.weak_bias2, new_weak_bias2)

            new_weak_variance  = self.weak_variance[self.iteration] + self.true_noise_level**2
            self.weak_variance = np.append(self.weak_variance, new_weak_variance)

            new_weak_risk = new_weak_bias2 + new_weak_variance
            self.weak_risk = np.append(self.weak_risk, new_weak_risk)

            new_strong_bias2  = self.strong_bias2[self.iteration] - self.true_signal[self.iteration]**2
            self.strong_bias2 = np.append(self.strong_bias2, new_strong_bias2)

            new_strong_variance  = self.strong_variance[self.iteration] + self.true_noise_level**2 / s**2
            self.strong_variance = np.append(self.strong_variance, new_strong_variance)

            new_strong_risk = new_strong_bias2 + new_strong_variance
            self.strong_risk = np.append(self.strong_risk, new_strong_risk)

        self.iteration += 1



import unittest
import numpy as np
from scipy.io import loadmat
import EarlyStopping as es


class TestConjugateGradientsGravity(unittest.TestCase):
    """Tests for the conjugate gradients algorithm"""

    def setUp(self):
        # setUp is a class from unittest

        # Python implementation of the gravity test problem from the regtools toolbox, see `Hansen (2008) <http://people.compute.dtu.dk/pcha/Regutools/RTv4manual.pdf>`_ for details
        self.sample_size = 2**9
        a = 0
        b = 1
        d = 0.25  # Parameter controlling the ill-posedness: the larger, the more ill-posed, default in regtools: d = 0.25

        t = (np.arange(1, self.sample_size + 1) - 0.5) / self.sample_size
        s = ((np.arange(1, self.sample_size + 1) - 0.5) * (b - a)) / self.sample_size
        T, S = np.meshgrid(t, s)

        self.design = (
            (1 / self.sample_size)
            * d
            * (d**2 * np.ones((self.sample_size, self.sample_size)) + (S - T) ** 2) ** (-(3 / 2))
        )
        self.signal = np.sin(np.pi * t) + 0.5 * np.sin(2 * np.pi * t)
        self.design_times_signal = self.design @ self.signal

        # Set parameters
        self.parameter_size = self.sample_size
        self.max_iter = self.sample_size
        self.noise_level = 10 ** (-2)

        # Specify number of Monte Carlo runs
        self.NUMBER_RUNS = 100

        # Create observations
        noise = np.random.normal(0, self.noise_level, (self.sample_size, self.NUMBER_RUNS))
        self.observation = noise + self.design_times_signal[:, None]

    def test_noise_free_model(self):
        # Test if conjugate gradient estimate converges to true signal in the noise free model
        model = es.ConjugateGradients(
            self.design,
            self.design @ self.signal,
            true_signal=self.signal,
            true_noise_level=0,
            computation_threshold=0,
        )
        model.iterate(2 * self.sample_size)
        self.assertAlmostEqual(sum((model.get_estimate(model.iteration) - self.signal) ** 2), 0, places=7)

    def calculate_residual(self, response, design, conjugate_gradient_estimate):
        return np.sum((response - design @ conjugate_gradient_estimate) ** 2)

    def test_residuals(self):
        # Test if the entry in the residuals vector at the noninterpolated discrepancy stopping index agrees with the squared residual of the conjugate gradient estimate at the same index
        critical_value = (np.sqrt(self.sample_size) + self.sample_size) * (self.noise_level**2)
        models = [
            es.ConjugateGradients(
                self.design,
                self.observation[:, i],
                true_signal=self.signal,
                true_noise_level=self.noise_level,
                computation_threshold=0,
            )
            for i in range(self.NUMBER_RUNS)
        ]

        max_iteration = self.sample_size
        for run in range(self.NUMBER_RUNS):
            early_stopping_index = models[run].get_discrepancy_stop(critical_value, max_iteration)
            if early_stopping_index is None:
                early_stopping_index = max_iteration
            residual = self.calculate_residual(
                models[run].response,
                models[run].design,
                models[run].get_estimate(early_stopping_index),
            )

            self.assertAlmostEqual(
                residual,
                models[run].residuals[int(early_stopping_index)],
                places=7,
            )

    def test_interpolation(self):
        # Test several properties of the interpolated conjugate gradients algorithm
        critical_value = (np.sqrt(self.sample_size) + self.sample_size) * (self.noise_level**2)
        models = [
            es.ConjugateGradients(
                self.design,
                self.observation[:, i],
                true_signal=self.signal,
                true_noise_level=self.noise_level,
                computation_threshold=0,
            )
            for i in range(self.NUMBER_RUNS)
        ]

        max_iteration = self.sample_size
        interpolation = True
        for run in range(self.NUMBER_RUNS):
            early_stopping_index = models[run].get_discrepancy_stop(critical_value, max_iteration, interpolation)
            if early_stopping_index is None:
                early_stopping_index = max_iteration
            interpolated_residual = models[run].get_residual(early_stopping_index)

            # Test if the interpolated squared residual at the discrepancy stopping index agrees with the critical value
            if early_stopping_index < max_iteration:
                self.assertAlmostEqual(interpolated_residual, critical_value, places=7)

            interpolated_residual_via_estimator = self.calculate_residual(
                models[run].response,
                models[run].design,
                models[run].get_estimate(early_stopping_index),
            )

            # Test if the interpolated squared residual at the discrepancy stopping index agrees with the squared residual of the conjugate gradient estimate at the same index
            self.assertAlmostEqual(interpolated_residual_via_estimator, interpolated_residual, places=7)

    def test_early_stopping_index(self):
        # Test if the discrepancy stopping index for the model without interpolation agrees with the rounded up discrepancy stopping index for the interpolated model
        critical_value = (np.sqrt(self.sample_size) + self.sample_size) * (self.noise_level**2)
        models = [
            es.ConjugateGradients(
                self.design,
                self.observation[:, i],
                true_signal=self.signal,
                true_noise_level=self.noise_level,
                computation_threshold=0,
            )
            for i in range(self.NUMBER_RUNS)
        ]

        max_iteration = self.sample_size
        for run in range(self.NUMBER_RUNS):
            early_stopping_index_interpolated = models[run].get_discrepancy_stop(critical_value, max_iteration, True)
            if early_stopping_index_interpolated is None:
                early_stopping_index_interpolated = max_iteration
            early_stopping_index_noninterpolated = models[run].get_discrepancy_stop(
                critical_value, max_iteration, False
            )
            if early_stopping_index_noninterpolated is None:
                early_stopping_index_noninterpolated = max_iteration
            self.assertAlmostEqual(
                np.ceil(early_stopping_index_interpolated),
                early_stopping_index_noninterpolated,
                places=7,
            )

    def test_empirical_oracles(self):
        # Test if the risks at the interpolated empirical oracles are smaller or equal to all risks along the integer iteration path
        # and, in addition, for supersmooth if they are smaller or equal to all risks along a real-valued grid of iteration indices

        models = [
            es.ConjugateGradients(
                self.design,
                self.observation[:, i],
                true_signal=self.signal,
                true_noise_level=self.noise_level,
                computation_threshold=0,
            )
            for i in range(self.NUMBER_RUNS)
        ]

        max_iteration = self.sample_size
        for run in range(self.NUMBER_RUNS):
            strong_empirical_oracle = models[run].get_strong_empirical_oracle(max_iteration, True)
            strong_empirical_oracle_risk = models[run].get_strong_empirical_risk(strong_empirical_oracle)

            self.assertTrue(all(strong_empirical_oracle_risk <= risk for risk in models[run].strong_empirical_risk))

            weak_empirical_oracle = models[run].get_weak_empirical_oracle(max_iteration, True)
            weak_empirical_oracle_risk = models[run].get_weak_empirical_risk(weak_empirical_oracle)

            self.assertTrue(all(weak_empirical_oracle_risk <= risk for risk in models[run].weak_empirical_risk))

            interpolated_strong_risk = []
            step_size = 0.1
            for iteration in np.arange(0, models[run].iteration + step_size, step_size):
                interpolated_strong_risk = np.append(
                    interpolated_strong_risk,
                    models[run].get_strong_empirical_risk(iteration),
                )

            self.assertTrue(all(strong_empirical_oracle_risk <= risk for risk in interpolated_strong_risk))

            interpolated_weak_risk = []
            step_size = 0.1
            for iteration in np.arange(0, models[run].iteration + step_size, step_size):
                interpolated_weak_risk = np.append(
                    interpolated_weak_risk,
                    models[run].get_weak_empirical_risk(iteration),
                )
            self.assertTrue(all(weak_empirical_oracle_risk <= risk for risk in interpolated_weak_risk))



from .example import add_one   # '.' needed since python3
from .L2_boost import L2_boost
from .conjugate_gradients import ConjugateGradients
from .landweber import Landweber
from .simulation_wrapper import SimulationWrapper
from .simulation_wrapper import SimulationParameters
from .simulation_wrapper import SimulationData
from .truncated_svd import TruncatedSVD
from .regression_tree import RegressionTree




import numpy as np
from sklearn import linear_model
import warnings

class L2_boost():
    """
    `[Source] <https://github.com/ESFIEP/EarlyStopping/blob/main/src/EarlyStopping/L2_boost.py>`_
    L2-boosting algorithm for high dimensional linear models.

    **Parameters**

    *design*: ``array``. nxp-Design matrix of the linear model.

    *response*: ``array``. n-dim vector of the observed data in the linear model.

    *true_signal*: ``array or None, default = None``. For simulation purposes only. For simulated data the true signal can be included to compute theoretical quantities such as the bias and the risk alongside the boosting procedure.

    **Attributes**

    *sample_size:* ``int``. Sample size of the linear model.
    
    *parameter_size*: ``int``. Parameter size of the linear model.

    *iteration*: ``int``. Current boosting iteration of the algorithm.

    *boost_estimate*: ``array``. Boosting estimate at the current iteration for the data given in design.

    *residuals*: ``array``. Lists the sequence of the residual mean of squares betwean the data and the boosting estimator.

    *bias2*: ``array``. Only exists if true_signal was given. Lists the values of the squared bias up to current boosting iteration.

    *stochastic_error*: ``array``. Only exists if true_signal was given. Lists the values of a stochastic error term up to current boosting iteration.

    *risk*: ``array``. Only exists if true_signal was given. Lists the values of the mean squared error between the boosting estimator and the true signal up to current boosting iteration.

    **Methods**

    +-----------------------------------------------------------------+------------------------------------------------------------------+
    | iterate( ``number_of_iterations=1`` )                           | Performs number of iterations of the boosting algorithm.         |
    +-----------------------------------------------------------------+------------------------------------------------------------------+
    | predict( ``design_observation`` )                               | Predicts the response based on the current boosting estimate.    |
    +-----------------------------------------------------------------+------------------------------------------------------------------+
    | get_discrepancy_stop( ``critical_value, max_iteration`` )       | Stops the boosting algorithm based on the discrepancy principle. | 
    +-----------------------------------------------------------------+------------------------------------------------------------------+
    | get_residual_ratio_stop( ``max_iteration, alpha=0.05, K=1.2`` ) | Stops the boosting algorithm based on the residual ratios.       | 
    +-----------------------------------------------------------------+------------------------------------------------------------------+
    | get_noise_estimate(``K=1``)                                     | Computes a noise estimate for the model via the scaled lasso.    |
    +-----------------------------------------------------------------+------------------------------------------------------------------+
    | get_aic_iteration(``K=2``)                                      | Computes the minimizer of a high dimensional Akaike criterion.   |
    +-----------------------------------------------------------------+------------------------------------------------------------------+
    | get_balanced_oracle( ``max_iteration`` )                        | Iterates the boosting algorithm up to the balanced oracle.       |
    +-----------------------------------------------------------------+------------------------------------------------------------------+
    """

    def __init__(self, design, response, true_signal = None):
        self.design      = design 
        self.response    = response
        self.true_signal = true_signal

        # Parameters of the model
        self.sample_size = np.shape(design)[0]
        self.parameter_size   = np.shape(design)[1]

        # Estimation quantities
        self.iteration           = 0
        self.selected_components = np.array([])
        self.orth_directions     = []
        self.coefficients_list   = [np.zeros(self.parameter_size)]
        self.boost_estimate_list = [np.zeros(self.sample_size)]

        # Residual quantities
        self.__residual_vector = response
        self.residuals         = np.array([np.mean(self.__residual_vector**2)])
        self.residual_ratios   = np.array([0])

        if self.true_signal is not None:
            self.noiseless_response   = self.design @ self.true_signal
            self.__error_vector       = self.response - self.noiseless_response
            self.__bias2_vector       = self.noiseless_response
            self.__stochastic_error_vector = np.zeros(self.sample_size)

            self.bias2      = np.array([np.mean(self.__bias2_vector**2)])
            self.stochastic_error = np.array([0])
            self.risk        = np.array([np.mean(self.noiseless_response**2)])

    def iterate(self, number_of_iterations = 1):
        """ Performs number_of_iterations iterations of the orthogonal boosting algorithm.

            **Parameters**

            *number_of_iterations*: ``int``. Number of boosting iterations to be performed.
        """
        for _ in range(number_of_iterations):
            self.__boost_one_iteration()

    def predict(self, design_observation):
        """ Predicts the output variable based on the current boosting estimate.

            **Parameters**

            *input_variable*: ``array``. The size of input_variable has to match parameter_size.
        """
        return np.dot(design_observation, self.coefficients)
        # 2025-03-10-TODO-BS: adjust coefficients.


    def get_discrepancy_stop(self, critical_value, max_iteration):
        """ Early stopping for the boosting procedure based on the discrepancy principle.
            Procedure is stopped when the residuals go below critical_value or iteration
            max_iteration is reached.

            **Parameters**

            *critical_value*: ``float``. Critical value for the early stopping procedure.

            *max_iteration*: ``int``. Maximal number of iterations to be performed.
        """
        if self.residuals[self.iteration] <= critical_value:
            # argmax takes the first instance of True in the true-false array
            early_stopping_index = np.argmax(self.residuals <= critical_value)
            return int(early_stopping_index)

        if self.residuals[self.iteration] > critical_value:
            while (self.residuals[self.iteration] > critical_value and
                   self.iteration <= max_iteration
                  ):
                self.iterate(1)

        if self.residuals[self.iteration] <= critical_value:
            early_stopping_index = self.iteration
            return int(early_stopping_index)
        else:
            warnings.warn("Stopping index not found up to max_iteration. Returning None.",
                          category=UserWarning)
            return None

    def get_residual_ratio_stop(self, max_iteration, alpha = 0.05, K = 1.2):
        """ Stops the algorithm based on a residual ration criterion.

            **Parameters**

            *alpha*: ``float``. accuracy level

            *K*: ``float``. Constant for the definition.

            *max_iteration*: ``int``. Maximal number of iterations to be performed.
        """ 
        critical_value = 1 - 4 * K * np.log(self.parameter_size / alpha) / self.sample_size
        if self.residual_ratios[self.iteration] >= critical_value:
            # argmax takes the first instance of True in the true-false array
            early_stopping_index = np.argmax(self.residual_ratios >= critical_value)
            return int(early_stopping_index)

        if self.residual_ratios[self.iteration] < critical_value:
            while (self.residual_ratios[self.iteration] < critical_value and
                   self.iteration <= max_iteration
                  ):
                self.iterate(1)

        if self.residual_ratios[self.iteration] >= critical_value:
            early_stopping_index = self.iteration
            return int(early_stopping_index)
        else:
            warnings.warn("Stopping index not found up to max_iteration. Returning None.",
                          category=UserWarning)
            return None

    def get_balanced_oracle(self, max_iteration):
        """ Performs iterations of the orthogonal boosting algorithm until the
            balanced oracle index at which the squared bias is smaller than the
            stochastic error is reached.

            **Parameters**

            *max_iteration*: ``int``. Maximal number of iterations to be performed.
        """
        if self.true_signal is None:
            return "This method is only available when the true signal is known"
        else:
            if self.bias2[self.iteration] <= self.stochastic_error[self.iteration]:
                # argmax takes the first instance of True in the true-false array
                balanced_oracle = np.argmax(self.bias2 <= self.stochastic_error)
                return int(balanced_oracle)

            if self.bias2[self.iteration] > self.stochastic_error[self.iteration]:
                while (self.bias2[self.iteration] > self.stochastic_error[self.iteration] and
                       self.iteration <= max_iteration
                      ):
                    self.iterate(1)

            if self.bias2[self.iteration] <= self.stochastic_error[self.iteration]:
                balanced_oracle = self.iteration
                return int(balanced_oracle)
            else:
                warnings.warn("Balanced oracle not found up to max_iteration. Returning None.",
                              category=UserWarning)
                return None

    def get_noise_estimate(self, K = 1):
        """ Computes an estimator for the noise level sigma^2 of the model via the scaled Lasso.

            **Parameters**

            *K*: ``float``. Constant in the definition. Defaults to 1, which is the choice from the scaled Lasso paper.
        """
        iter = 0
        max_iter = 50
        tolerance = 1 / self.sample_size
        estimation_difference = 2 * tolerance
        alpha_0 = np.sqrt(K * np.log(self.parameter_size) / self.sample_size)

        lasso = linear_model.Lasso(alpha_0, fit_intercept = False)
        lasso.fit(self.design, self.response)
        noise_estimate = np.mean((self.response - lasso.predict(self.design))**2)

        while estimation_difference > tolerance and iter <= max_iter:
            alpha = np.sqrt(noise_estimate) * alpha_0
            lasso = linear_model.Lasso(alpha, fit_intercept = False)
            lasso.fit(self.design, self.response)

            new_noise_estimate = np.mean((self.response - lasso.predict(self.design))**2)
            estimation_difference = np.abs(new_noise_estimate - noise_estimate)
            noise_estimate = new_noise_estimate

            iter = iter + 1

        return noise_estimate

    def get_aic_iteration(self, K = 2, max_iteration=None):
        """ Computes the iteration index minimizing a high dimensional Akaike criterion.

            **Parameters**

            *K*: ``float``. Constant in the definition. Defaults to 2, which is common in the literature.
        """
        if (max_iteration == None):
            max_iteration = self.iteration
            
        noise_estimate = self.get_noise_estimate()
        dim_penalty = np.arange(0, max_iteration + 1) * K * noise_estimate * np.log(self.parameter_size) / self.sample_size
        aic = self.residuals[0:max_iteration + 1] + dim_penalty
        return np.argmin(aic)


    def __boost_one_iteration(self):
        """Performs one iteration of the orthogonal boosting algorithm"""
        # Compute weak learner index and check for repetition
        weak_learner_index            = self.__compute_weak_learner_index()
        component_selected_repeatedly = False
        for iter_num in range(self.iteration):
            if weak_learner_index == self.selected_components[iter_num]:
                component_selected_repeatedly = True

        if component_selected_repeatedly:
            print("Algorithm terminated")
        else:
            # Update selected variables
            self.selected_components = np.append(self.selected_components, weak_learner_index)
            self.__update_orth_directions(self.design[:, weak_learner_index])
            weak_learner = self.orth_directions[-1]

            # Update estimation quantities
            coefficient_entry                    = np.dot(self.response, weak_learner) / self.sample_size
            new_coefficients                     = self.coefficients_list[self.iteration]
            new_coefficients[weak_learner_index] = coefficient_entry
            self.coefficients_list.append(new_coefficients)

            new_boost_estimate = self.boost_estimate_list[self.iteration] + coefficient_entry * weak_learner
            self.boost_estimate_list.append(new_boost_estimate)

            self.__residual_vector = self.response - new_boost_estimate
            new_residuals          = np.mean(self.__residual_vector**2)
            new_residual_ratio     = new_residuals / self.residuals[self.iteration]
            self.residuals         = np.append(self.residuals, new_residuals)
            self.residual_ratios   = np.append(self.residual_ratios, new_residual_ratio)

            self.iteration              = self.iteration + 1

            # Update theoretical quantities
            if self.true_signal is not None:
                self.__update_risk()
                self.__update_bias2(weak_learner)
                self.__update_stochastic_error(weak_learner)

    def __compute_weak_learner_index(self):
        """ Computes the column index of the design matrix which reduces the
            resiudals the most
        """
        decreased_residuals = np.zeros(self.parameter_size)
        for j in range(self.parameter_size):
            direction             = self.design[:, j]
            direction_norm         = np.sqrt(np.mean(direction**2))
            direction             = direction / direction_norm
            step_size              = np.dot(self.__residual_vector, direction) / \
                                    self.sample_size
            decreased_residuals[j] = np.mean((self.__residual_vector -
                                             step_size * direction)**2)
        weak_learner_index = np.argmin(decreased_residuals)
        return weak_learner_index

    def __update_orth_directions(self, direction):
        """Updates the list of orthogonal directions"""
        if self.iteration == 0:
            direction_norm = np.sqrt(np.mean(direction**2))
            direction     = direction / direction_norm
        else:
            for orth_direction in self.orth_directions:
                dot_product = np.dot(direction, orth_direction) / self.sample_size
                direction  = direction -  dot_product * orth_direction
            direction_norm = np.sqrt(np.mean(direction**2))
            direction     = direction / direction_norm
        self.orth_directions.append(direction)

    def __update_risk(self):
        new_risk   = np.mean((self.noiseless_response - self.boost_estimate_list[self.iteration])**2)
        self.risk = np.append(self.risk, new_risk)

    def __update_bias2(self, weak_learner):
        coefficient        = np.dot(self.noiseless_response, weak_learner) / \
                             self.sample_size
        self.__bias2_vector = self.__bias2_vector - coefficient * weak_learner
        new_bias2           = np.mean(self.__bias2_vector**2)
        self.bias2         = np.append(self.bias2, new_bias2)

    def __update_stochastic_error(self, weak_learner):
        coefficient             = np.dot(self.__error_vector, weak_learner) / \
                                 self.sample_size
        self.__stochastic_error_vector = self.__stochastic_error_vector + \
                                  coefficient * weak_learner
        new_stochastic_error           = np.mean(self.__stochastic_error_vector**2)
        self.stochastic_error         = np.append(self.stochastic_error, new_stochastic_error)



from __future__ import annotations
import numpy as np
import pandas as pd
from queue import Queue
import warnings

class RegressionTree:
    """
    A class for constructing a regression tree and computing theoretical quantities.

    **Parameters**

    *design*: ``array``. The design matrix for the regression tree.

    *response*: ``array``. The response variable values.

    *min_samples_split*: ``int``. The number of samples that the terminal node can have at maximum.

    *true_signal*: ``array or None``. Used only in simulation contexts for computing theoretical quantities.

    *true_noise_vector*: ``array or None``. Used only in simulation contexts for theoretical quantities.

    **Attributes**

    *sample_size*: ``int``. The number of samples in the data.

    *dimension*: ``int``. The number of variables in the design matrix.

    *regression_tree*: ``Node``. The root node of the regression tree.

    *residuals*: ``array``. Stores the mean squared error residuals at each level of the tree.

    *bias2*: ``array``. Stores the squared bias values at each level of the tree.

    *variance*: ``array``. Stores the variance values at each level of the tree.

    *risk*: ``array``. Risk based on bias and variance.

    **Methods**

    +-------------------------------------------+--------------------------------------------------------------------+
    | iterate( ``max_depth=None`` )             | Grows a regression tree up to the specified depth.                 |
    +-------------------------------------------+--------------------------------------------------------------------+
    | predict( ``design, depth`` )              | Predicts target values using the regression tree at a given depth. |
    +-------------------------------------------+--------------------------------------------------------------------+
    | get_discrepancy_stop( ``crit, max_depth``)| Finds the first generation satisfying the discrepancy principle.   |
    +-------------------------------------------+--------------------------------------------------------------------+
    | get_balanced_oracle( ``max_depth`` )      | Computes the balanced oracle generation.                           |
    +-------------------------------------------+--------------------------------------------------------------------+
    """

    class Node:
        """
        A class representing a node of the regression tree.
        """
        def __init__(self):
            self.split_threshold = None
            self.variable = None
            self.left_child = None
            self.right_child = None
            self.is_terminal = False
            self.node_prediction = None

        def set_params(self, split_threshold: float, variable: int) -> None:
            """ Set splitting observation and splitting variable for this node. """
            self.split_threshold = split_threshold
            self.variable = variable

    def __init__(self,
                 design: np.array,
                 response: np.array,
                 min_samples_split,
                 true_signal = None,
                 true_noise_vector = None):

        self.true_signal = true_signal
        self.minimal_samples_split = min_samples_split
        self.true_noise_vector = true_noise_vector
        self.design = design
        self.response = response
        self.sample_size = self.design.shape[0]
        self.dimension = self.design.shape[1]



    def iterate(self, max_depth: int = None):
        """
        Grows the regression tree up to the specified depth.

        **Parameters**

        *max_depth*: ``int or None``. Maximum depth to which the tree should grow. If None, the tree grows fully.
        """

        self.residuals = np.array([])
        # For theoretical quantities:
        self.block_matrix = {}
        self.indices_processed = {}
        self.block_matrices_full = {}
        self.bias2 = np.array([])
        self.variance = np.array([])
        self.risk = np.array([])

        # Initialize root node of the tree
        self.regression_tree = self.Node()
        self.queue = Queue()
        self.queue.put((self.regression_tree, np.arange(self.sample_size), 1)) # start with level 1
        self.terminal_indices = {}

        # Process all nodes at the current level (= perform 'one iteration'):
        self.maximal_depth = max_depth
        while not self.queue.empty():
            self._grow_one_iteration()

    def _grow_one_iteration(self):
        """
        Grows one iteration of the regression tree using a breadth-first approach.
        """

        #Initialize:
        level_mse_sum = 0
        level_node_count = 0
        next_level_queue = Queue()
        self.level_indices = []
        current_level_observations = {}

        # Process all nodes at the current level
        for _ in range(self.queue.qsize()):
            self.node, parent_indices, level = self.queue.get()
            number_observations_node = len(self.response[parent_indices])
            self.node.node_prediction = np.mean(self.response[parent_indices])

            # Calculate MSE for the current node and update level MSE
            level_mse_sum += self._impurity(parent_indices) * (number_observations_node / self.sample_size)
            level_node_count += 1

            # Check termination conditions
            terminal_due_to_samples = number_observations_node == self.minimal_samples_split
            terminal_due_to_depth = self.maximal_depth is not None and level >= self.maximal_depth

            # If the terminal condition is satisfied, the node is not added to the next level queue.
            if terminal_due_to_samples or terminal_due_to_depth:
                self.node.is_terminal = True
                continue

            else:
                split_variable, split_value, left_indices, right_indices = self._find_best_split_parent(indices=parent_indices)

                # ILeft and right child node
                self.node.left_child = self.Node()
                self.node.right_child = self.Node()
                self.node.left_child.node_prediction = np.mean(self.response[left_indices])
                self.node.right_child.node_prediction = np.mean(self.response[right_indices])
                self.node.set_params(split_value, split_variable)
                self.level_indices.extend([left_indices, right_indices])

                # Add child nodes to the next level queue
                next_level_queue.put((self.node.left_child, left_indices, level + 1))
                next_level_queue.put((self.node.right_child, right_indices, level + 1))

                # Store observations for each node
                current_level_observations[self.node.left_child] = len(self.response[left_indices])
                current_level_observations[self.node.right_child] = len(self.response[right_indices])

                if len(left_indices) == self.minimal_samples_split:
                    if level not in self.terminal_indices:
                        self.terminal_indices[level] = []
                    self.terminal_indices[level].append(left_indices)

                if len(right_indices) == self.minimal_samples_split:
                    if level not in self.terminal_indices:
                        self.terminal_indices[level] = []
                    self.terminal_indices[level].append(right_indices)


        # Update the residuals, theoretical quantities, and queue for the next level
        self.residuals = np.append(self.residuals, level_mse_sum)
        self._update_theoretical_quantities(current_level_observations, level)
        self.queue = next_level_queue


    def _find_best_split_parent(self, indices):
        """
        Finds the best split based on impurity reduction.

        **Parameters**

        *indices*: ``list``. Indices of samples to be split.
        """
        best_impurity = float("inf")
        best_split = (None, None, [], [])
        range_variables = range(self.dimension)

        # Iterate through the possible variable/split combinations
        for variable in range_variables:
            thresholds = np.unique(self.design[indices, variable])
            for split_threshold in thresholds:
                left_node_indices = [i for i in indices if self.design[i, variable] <= split_threshold]
                right_node_indices = [i for i in indices if self.design[i, variable] > split_threshold]

                if len(left_node_indices) == 0 or len(right_node_indices) == 0:
                    continue
                impurity = (
                            len(self.design[left_node_indices]) * self._impurity(left_node_indices)
                            + len(self.design[right_node_indices]) * self._impurity(right_node_indices)
                           ) / self.sample_size

                # Update the impurity and choice of variable/split
                if impurity < best_impurity:
                    best_impurity = impurity
                    best_split = (variable, split_threshold, np.array(left_node_indices), np.array(right_node_indices))

        return best_split

    def predict(self, design: pd.DataFrame | np.ndarray, depth: int):
        """
        Predicts target values for the given design data using the decision tree at the specified depth.

        **Parameters**

        *design*: ``array``. Input design matrix for predictions.

        *depth*: ``int``. Depth level of the tree to use for predictions. If 0, returns the unconditional mean.
        """

        if isinstance(design, pd.DataFrame):
            design = design.to_numpy()

        # Handle depth = 0 case by returning the unconditional mean
        if depth == 0:
            return np.repeat(np.mean(self.response), design.shape[0])
        if depth > self.maximal_depth:
            warnings.warn("Depth can not exceed max_depth. Returning None.", category=UserWarning)
            return None

        predictions = []
        for sample in design:
            current_node = self.regression_tree  # Start at the root
            current_depth = 1

            # Traverse the tree based on stored nodes up to the specified depth
            while not current_node.is_terminal and current_depth <= depth:
                # Find the current node's split criteria based on the level structure
                split_feature = current_node.variable
                split_threshold = current_node.split_threshold

                # Decide which child node to move to based on split criteria
                if sample[split_feature] <= split_threshold:
                    current_node = current_node.left_child
                else:
                    current_node = current_node.right_child

                current_depth += 1

            # When reaching the terminal node or max depth, take the prediction value
            predictions.append(current_node.node_prediction)

        return np.array(predictions)

    def _update_theoretical_quantities(self, current_level_observations, level):
        """
        Updates theoretical quantities (bias, variance, risk) at each level.

        **Parameters**

        *current_level_observations*: ``dict``. Observations at the current tree level.

        *level*: ``int``. Current level of the tree.
        """
        # Exit the function if empty
        if not self.level_indices:
            return

        # Processing of block matrix:
        self._block_matrix_processing(current_level_observations, level)
        if self.true_signal is not None and self.true_noise_vector is not None:
            if self.block_matrix[level].shape[0] == self.sample_size:
                self.new_bias2, self.new_variance = self._get_bias_and_variance(self.indices_processed[level],
                                                                                self.block_matrix[level])
                self.bias2 = np.append(self.bias2, self.new_bias2)
                self.variance = np.append(self.variance, self.new_variance)
                self.new_risk = self.new_bias2 + self.new_variance
                self.risk = np.append(self.risk, self.new_risk)

            else:
                self.new_bias2, self.new_variance = self._get_bias_and_variance(self.indices_complete,
                                                                                self.block_matrices_full[level])
                self.bias2 = np.append(self.bias2, self.new_bias2)
                self.variance = np.append(self.variance, self.new_variance)
                self.new_risk = self.new_bias2 + self.new_variance
                self.risk = np.append(self.risk, self.new_risk)

    def get_discrepancy_stop(self, critical_value, max_depth = None):
        """
         Finds the first generation where the discrepancy principle is met.

         **Parameters**

         *critical_value*: ``float``. Threshold for discrepancy-based stopping.

         *max_depth*: ``None``. Maximum depth for the tree if it has not been grown yet.
         """

        # If no iteration done before, grow the tree until max_depth
        if self.residuals.size == 0:
            self.maximal_depth = max_depth
            while not self.queue.empty():
                self._grow_one_iteration()

        if np.any(self.residuals <= critical_value):
            # argmax takes the first instance of True in the true-false array
            early_stopping_index = np.argmax(self.residuals <= critical_value)
            return early_stopping_index
        else:
            warnings.warn("Early stopping index not found. Returning None.", category=UserWarning)
            return None

    def get_balanced_oracle(self, max_depth = None):
        """
        Computes the balanced oracle iteration based on the bias and variance.

        **Parameters**

        *max_depth*: ``None``. Maximum depth for the tree if it has not been grown yet.
        """

        # If no iteration done before, grow the tree until max_depth
        if self.residuals.size == 0:
            self.maximal_depth = max_depth
            while not self.queue.empty():
                self._grow_one_iteration()

        if np.any(self.bias2 <= self.variance):
            # argmax takes the first instance of True in the true-false array
            balanced_oracle = np.argmax(self.bias2 <= self.variance)
            return balanced_oracle

        else:
            warnings.warn(
                "Balanced oracle not found. Returning None.", category=UserWarning
            )
            return None

    def _block_matrix_processing(self, current_level_observations, level):
        # Process observations and indices after completing the level
        self.block_matrix[level] = self._process_level_observations(current_level_observations)
        self.indices_processed[level] = np.concatenate(self.level_indices)

        if self.block_matrix[level].shape[0] < self.sample_size:
            indices_pre_append = self.indices_processed[level]
            filtered_indices = {k: v for k, v in self.terminal_indices.items() if k != level}

            # Check if filtered_indices is None or empty
            if not filtered_indices:
                print("No indices to concatenate.")
                return

            self.block_matrices_full[level] = self._append_block_matrix(self.block_matrix[level],
                                                                            filtered_indices)
            indices_append = np.concatenate(
                [idx for level in range(1, level) for idx in self.terminal_indices.get(level, [])])
            self.indices_complete = np.append(indices_pre_append, indices_append)

    def _impurity(self, indices) -> float:
        """
        Computes the mean squared error impurity for the response at the given set of indices.

        **Parameters**

        *indices*: ``list``. Indices of samples to compute impurity on.
        """
        response_node = self.response[indices]
        mse = np.sum((response_node -  np.mean(response_node)) ** 2) / len(response_node)
        return mse

    def _append_block_matrix(self, existing_matrix, filtered_indices):
        """
            Appends new block matrices to an existing block-diagonal
            matrix to create an expanded block-diagonal matrix.

            This method constructs a larger block-diagonal matrix by appending
            new blocks derived from `filtered_indices`
            to an `existing_matrix`. Each new block corresponds to the size of
            index arrays provided in `filtered_indices`
            and contains entries that are the reciprocal of the block size (1/size).

            Parameters:
                existing_matrix (np.ndarray or None): The existing block-diagonal matrix
                to which new blocks will be appended.
                    - If `None` or an empty array, a new block-diagonal matrix is created from the new blocks alone.
                filtered_indices (dict): A dictionary where each key corresponds to a level
                or identifier, and each value
                    is a list of NumPy arrays. Each array represents indices of data points, and its size determines
                    the dimensions of the corresponding block matrix.

            Returns:
                np.ndarray: A new block-diagonal matrix that includes the existing matrix and the newly appended blocks.

            """
        elements_count = {key: [arr.size for arr in value] for key, value in filtered_indices.items()}

        # Collect all block sizes and create each block
        block_matrices = []
        total_new_block_size = 0

        # Iterate over each key-value pair to create each block matrix
        for sizes in elements_count.values():
            for size in sizes:
                # Create a block of size 'size x size' with entries 1/size
                block_matrix = np.full((size, size), 1 / size, dtype=float)
                block_matrices.append(block_matrix)
                total_new_block_size += size

        # Check if there is anything to append
        if total_new_block_size == 0:
            return existing_matrix

        # Create a large matrix that will include the existing and all new blocks
        if existing_matrix is None or existing_matrix.size == 0:
            # If no existing matrix, simply concatenate all new blocks
            new_block_matrix = block_matrices[0]
            for block in block_matrices[1:]:
                new_block_matrix = np.block([
                    [new_block_matrix, np.zeros((new_block_matrix.shape[0], block.shape[1]))],
                    [np.zeros((block.shape[0], new_block_matrix.shape[1])), block]
                ])
            return new_block_matrix
        else:
            # Existing matrix is present; calculate its dimension
            existing_dim = existing_matrix.shape[0]
            # Create a full matrix to hold both existing and new blocks
            full_matrix = np.zeros((existing_dim + total_new_block_size, existing_dim + total_new_block_size))
            full_matrix[:existing_dim, :existing_dim] = existing_matrix

            # Place new blocks starting from the bottom-right of the existing matrix
            start_dim = existing_dim
            for block in block_matrices:
                end_dim = start_dim + block.shape[0]
                full_matrix[start_dim:end_dim, start_dim:end_dim] = block
                start_dim = end_dim  # Update the starting dimension for the next block

            return full_matrix

    def _get_bias_and_variance(self, indices, block_matrix):
        """
           Calculates the bias squared and variance for a given iteration level to
           determine the balanced oracle iteration.

           Parameters:
               indices (np.ndarray): An array of indices corresponding to the data points being considered.
               block_matrix (np.ndarray): A matrix used in the computation of bias and variance.
           Returns:
               tuple:
                   balanced_oracle_iteration (int or None): The current `level` if `bias2` is less than
                   or equal to `variance`;
                       otherwise, `None`.
                   bias2 (float): The computed bias squared value.
                   variance (float): The computed variance value.

           """

        # Squared Bias:
        bias2 = np.mean(((np.eye(indices.shape[0]) - block_matrix) @ self.true_signal[indices]) ** 2)
        # Variance:
        variance = np.mean((block_matrix @ self.true_noise_vector[indices]) ** 2)

        return bias2, variance

    def _process_level_observations(self, observations):
        """
            Constructs a block-diagonal matrix from observations at a specific level of the tree.

            This method processes the observations collected from nodes at a particular level during
            the tree-growing process and creates a block-diagonal matrix. Each block corresponds to
            a node and is a square matrix of size equal to the number of observations at that node.

            Parameters:
                observations (dict): A dictionary where each key is a node object, and each value is
                    an integer representing the number of observations at that node.

            Returns:
                np.ndarray or None:
                    - If observations are provided and valid blocks are created, returns a NumPy ndarray
                      representing the assembled block-diagonal matrix.
                    - If there are no observations (empty dictionary), returns `None`.
                    - If no valid block matrices are created (e.g., all counts are zero), returns an empty array.

            """


        block_matrices = []
        dimensions = []

        # Iterate over each node's observations to create individual block matrices
        for node, count in observations.items():
            if count > 0:
                entry = 1 / count  #
                block_matrix = np.full((count, count), entry)  # Create a block matrix for the node
                block_matrices.append(block_matrix)
                dimensions.append(count)

        # Combine the block matrices into a single large block matrix
        total_dim = sum(dimensions)
        full_matrix = np.zeros((total_dim, total_dim))

        # Populate the full matrix with individual block matrices
        current_position = 0
        for i, block in enumerate(block_matrices):
            dim = dimensions[i]
            full_matrix[current_position:current_position + dim, current_position:current_position + dim] = block
            current_position += dim

        return full_matrix






import unittest
import numpy as np
import EarlyStopping as es


class Test_landweber(unittest.TestCase):
    """Tests for the landweber algorithm"""

    def setUp(self):
        self.design = np.diag([1, 2, 3, 4, 5])
        self.true_signal = np.random.normal(5, 1, 5)
        self.tol = 10 ** (-7)
        self.response = self.design @ self.true_signal

    def test_matrix_inversion(self):
        self.alg = es.Landweber(self.design, self.response, true_signal=self.true_signal, learning_rate=0.01)
        landweber_estimate = self.alg.get_estimate(500)
        self.assertAlmostEqual(np.sum((landweber_estimate - self.true_signal) ** 2), 0, places=2)

    def test_discrepancy_stop(self):
        self.alg = es.Landweber(self.design, self.response, true_signal=self.true_signal, learning_rate=0.01)
        critical_value = 0.001
        early_stopping_index = self.alg.get_discrepancy_stop(critical_value, max_iteration=1000)
        residual_condition_1 = self.alg.residuals[early_stopping_index - 1] > critical_value
        residual_condition_2 = self.alg.residuals[early_stopping_index] <= critical_value

        self.assertTrue(residual_condition_1 and residual_condition_2)

        landweber_estimate = self.alg.get_estimate(early_stopping_index)
        self.assertAlmostEqual(np.sum((landweber_estimate - self.true_signal) ** 2), 0, places=2)

    def test_weak_oracle(self):
        self.response = self.response + np.random.normal(0, 0.01, 5)
        self.alg = es.Landweber(
            self.design, self.response, true_signal=self.true_signal, learning_rate=0.01, true_noise_level=0.01
        )

        weak_balanced_oracle = self.alg.get_weak_balanced_oracle(1000)
        weak_oracle_condition_1 = (
            self.alg.weak_bias2[weak_balanced_oracle - 1] > self.alg.weak_variance[weak_balanced_oracle - 1]
        )
        weak_oracle_condition_2 = (
            self.alg.weak_bias2[weak_balanced_oracle] <= self.alg.weak_variance[weak_balanced_oracle]
        )

        self.assertTrue(weak_oracle_condition_1 and weak_oracle_condition_2)

    def test_strong_oracle(self):
        self.response = self.response + np.random.normal(0, 0.01, 5)
        self.alg = es.Landweber(
            self.design, self.response, true_signal=self.true_signal, learning_rate=0.01, true_noise_level=0.01
        )

        strong_balanced_oracle = self.alg.get_strong_balanced_oracle(1000)
        strong_oracle_condition_1 = (
            self.alg.strong_bias2[strong_balanced_oracle - 1] > self.alg.strong_variance[strong_balanced_oracle - 1]
        )
        strong_oracle_condition_2 = (
            self.alg.strong_bias2[strong_balanced_oracle] <= self.alg.strong_variance[strong_balanced_oracle]
        )

        self.assertTrue(strong_oracle_condition_1 and strong_oracle_condition_2)


# Running the tests
if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
import EarlyStopping as es
from scipy.sparse.linalg import svds

class TestTruncatedSVD(unittest.TestCase):
    def setUp(self):
        # setUp is a class from unittest
        self.sample_size = 5
        self.para_size = 5

    def test_inversion_without_noise(self):
        design = np.random.normal(0, 1, size = (self.sample_size, self.sample_size))
        signal = np.random.uniform(0, 1, size = self.sample_size)
        noiseless_response = design @ signal
        alg = es.TruncatedSVD(design, noiseless_response)
        alg.iterate(self.sample_size)
        truncated_svd_estimate = alg.get_estimate(alg.iteration)
        self.assertAlmostEqual(np.mean(truncated_svd_estimate - signal), 0, places=5)

    def test_diagonal_inversion_without_noise(self):
        design = np.diag(np.random.uniform(0, 1, size = self.sample_size))
        signal = np.random.uniform(0, 1, size = self.sample_size)
        noiseless_response = design @ signal
        alg = es.TruncatedSVD(design, noiseless_response, diagonal = True)
        alg.iterate(self.sample_size)
        truncated_svd_estimate = alg.get_estimate(alg.iteration)
        self.assertAlmostEqual(np.mean(truncated_svd_estimate - signal), 0, places=5)

    def test_monotonicity_of_theoretical_quantities(self):
        design = np.random.normal(0, 1, size = (self.sample_size, self.sample_size))
        signal = np.random.uniform(0, 1, size = self.sample_size)
        response = design @ signal + np.random.normal(0, 0.1, self.sample_size)
        alg = es.TruncatedSVD(design, response, true_signal = signal, true_noise_level = 0.1)

        for _ in range(self.sample_size):
            alg.iterate(1)

            self.assertLessEqual(alg.weak_bias2[alg.iteration], alg.weak_bias2[alg.iteration - 1])
            self.assertLessEqual(alg.strong_bias2[alg.iteration], alg.strong_bias2[alg.iteration - 1])

            self.assertLessEqual(alg.weak_variance[alg.iteration - 1], alg.weak_variance[alg.iteration])
            self.assertLessEqual(alg.strong_variance[alg.iteration - 1], alg.strong_variance[alg.iteration])

    def test_diagonal_monotonicity_of_theoretical_quantities(self):
        design = np.diag(np.random.uniform(0, 1, size = self.sample_size))
        signal = np.random.uniform(0, 1, size = self.sample_size)
        response = design @ signal + np.random.normal(0, 0.1, self.sample_size)
        alg = es.TruncatedSVD(design, response, true_signal = signal, true_noise_level = 0.1, diagonal = True)

        for _ in range(self.sample_size):
            alg.iterate(1)

            self.assertLessEqual(alg.weak_bias2[alg.iteration], alg.weak_bias2[alg.iteration - 1])
            self.assertLessEqual(alg.strong_bias2[alg.iteration], alg.strong_bias2[alg.iteration - 1])

            self.assertLessEqual(alg.weak_variance[alg.iteration - 1], alg.weak_variance[alg.iteration])
            self.assertLessEqual(alg.strong_variance[alg.iteration - 1], alg.strong_variance[alg.iteration])


if __name__ == '__main__':
     unittest.main()



import unittest
import numpy as np
from EarlyStopping import L2_boost

class Test_L2_boost(unittest.TestCase):
    """Tests for the L2_boost algorithm"""

    def setUp(self):
        # setUp is a class from unittest
        # Simulate data
        self.sample_size = 5
        self.para_size   = 5
        self.X           = np.random.normal(0, 1, size = (self.sample_size, self.para_size))
        self.f           = 15 * self.X[:, 0] + 10 * self.X[:, 1] + 5 * self.X[:, 2]
        self.eps         = np.random.normal(0, 5, self.sample_size)
        self.Y           = self.f + self.eps
        self.tol         = 10**(-5)

    def test_termination_of_the_algorithm(self):
        self.alg = L2_boost(self.X, self.Y)
        self.alg.iterate(self.alg.sample_size + 1)
        self.assertTrue(self.alg.iter < self.alg.sample_size + 1)

    def test_orthonormalization(self):
        self.alg = L2_boost(self.X, self.Y)
        self.alg.iterate(self.alg.sample_size)
        for m in range(self.alg.iter):
            direction_m = self.alg.orth_directions[m]
            deviation_vector = np.zeros(self.alg.sample_size)
            for j in range(self.alg.iter):
                direction_j = self.alg.orth_directions[j]
                if j == m:
                    deviation_vector[j] = 1 - np.mean(direction_j**2)
                else:
                    deviation_vector[j] = np.dot(direction_m, direction_j) / self.alg.sample_size
            condition_vector = ( np.absolute(deviation_vector) > self.tol )
            number_of_deviations_larger_tol = np.sum(condition_vector)
            self.assertTrue(number_of_deviations_larger_tol == 0)

    def test_monotonicity_of_bias_and_variance(self):
        self.alg = L2_boost(self.X, self.Y, true_signal = self.f)
        self.alg.iterate(self.sample_size)
        for m in range((self.alg.iter - 1)):
            if self.alg.bias2[m] >= self.tol:
                self.assertTrue(self.alg.bias2[m] >= self.alg.bias2[(m + 1)])
            self.assertTrue(self.alg.stoch_error[m] <= self.alg.stoch_error[(m + 1)])

    def test_consistency_of_bias_variance_computation(self):
        self.alg = L2_boost(self.X, self.Y, true_signal = self.f)
        self.alg.iterate(self.alg.sample_size)
        alternative_computation_mse = self.alg.bias2 + self.alg.stoch_error
        deviation_vector = np.abs(alternative_computation_mse - self.alg.mse)
        for m in range(self.alg.iter):
            self.assertTrue(deviation_vector[m] < self.tol)

    def test_limit_of_the_stochastic_error(self):
        self.alg = L2_boost(self.X, self.Y, true_signal = self.f)
        self.alg.iterate(self.alg.sample_size)
        avg_squared_error = np.mean(self.eps**2)
        last_index = self.alg.iter
        deviation = np.abs(avg_squared_error - self.alg.stoch_error[last_index])
        self.assertTrue(deviation < self.tol)

    def test_boost_to_balanced_oracle(self):
        self.alg = L2_boost(self.X, self.Y, true_signal = self.f)
        self.alg.boost_to_balanced_oracle()
        self.assertTrue(self.alg.bias2[self.alg.iter] <= self.alg.stoch_error[self.alg.iter])
        self.assertTrue(self.alg.bias2[(self.alg.iter - 1)] > self.alg.stoch_error[(self.alg.iter -1)])

# if __name__ == '__main__':
#     unittest.main()



import numpy as np
from joblib import Parallel, delayed
from scipy.linalg import toeplitz
import scipy
from scipy import sparse
from scipy.sparse import dia_matrix
from scipy.sparse.linalg import svds
import pandas as pd
import warnings

from .landweber import Landweber
from .conjugate_gradients import ConjugateGradients
from .truncated_svd import TruncatedSVD
from .L2_boost import L2_boost

# import matplotlib.pyplot as plt
# from matplotlib.ticker import FixedLocator
# import seaborn as sns
# import pandas as pd
# TODO: Update bias2 vs. bias/ mse vs risk, landweber empirical vs. non-empirical quantities for replication study
# bias2
# risk


def custom_warning(message, category, filename, lineno, file=None, line=None):
    message_str = str(message)  # Convert to string for checking
    if "PARAMETER WARNING" in message_str:  # Only color specific warnings
        return f"\033[91m{category.__name__}: {message}\033[0m\n"  # Red
    else:
        return f"{category.__name__}: {message}\n"  # Default color


warnings.formatwarning = custom_warning


class SimulationData:
    """
     `[Source] <https://github.com/ESFIEP/EarlyStopping/edit/main/src/EarlyStopping/simulation_wrapper.py>`_ A collection of static methods for the creation of simmulation data.

    **Description**

    Collection of serveral important examples of inverse problems.

    **References**

     `[Toolbox] <https://www.mathworks.com/matlabcentral/fileexchange/52-regtools>`_ The ill-posed inverse problems heat, deriv2, gravity, phillips are based on the Hansen-Matlab toolbox.

    **Methods**

    +----------------------------------------------------+----------------------------------------------------------------------------------------------+
    | diagonal_data(``sample_size``, ``type``)           | Creation of diagonal design with `smooth`, `supersmooth` and `rough` signals                 |
    +----------------------------------------------------+----------------------------------------------------------------------------------------------+
    | gravity(``sample_size``, ``a``, ``b``, ``d``)      | Discretised gravity operator with parameters controlling the ill-posedness  of the problem   |
    +----------------------------------------------------+----------------------------------------------------------------------------------------------+
    | heat(``sample_size``, ``kappa``)                   | Discretised heat semigroup with parameter controlling the ill-conditioning of the matrix     |
    +----------------------------------------------------+----------------------------------------------------------------------------------------------+
    | deriv2(``sample_size``, ``example``)               | Discretisation of a Fredholm integral equation whose kernel K is the Green's function        |
    +----------------------------------------------------+----------------------------------------------------------------------------------------------+
    | phillips(``sample_size``)                          | Create data based on the famous Phillips example                                             |
    +----------------------------------------------------+----------------------------------------------------------------------------------------------+
    | additive_smooth(``sample_size``, ``noise_level``)  | Create data with a smooth additive model using sine, quadratic, linear, and exp.components   |
    +----------------------------------------------------+----------------------------------------------------------------------------------------------+
    | additive_step(``sample_size``, ``noise_level``)    | Create data with a stepwise additive model using piecewise constant functions                |
    +----------------------------------------------------+----------------------------------------------------------------------------------------------+
    | additive_linear(``sample_size``, ``noise_level``)  | Create data with a linear additive model using piecewise linear interpolation                |
    +----------------------------------------------------+----------------------------------------------------------------------------------------------+
    | additive_hills(``sample_size``, ``noise_level``)   | Create data with a hills additive model using sinusoidal components with varying frequencies |
    +----------------------------------------------------+----------------------------------------------------------------------------------------------+
    """


    @staticmethod
    def additive_smooth(sample_size, noise_level):

        noise = np.random.normal(0, noise_level, sample_size)
        design = np.random.uniform(-2.5, 2.5, size=(sample_size, 30))

        true_signal = -2 * np.sin(2 * design[:,0]) + (0.8 * design[:,1]**2 - 2.5) + (design[:,2] - 1/2) + (np.exp(-0.65 * design[:,3]) - 2.5)
        response = true_signal + noise

        return design, response, true_signal
    
    @staticmethod   
    def additive_hills(sample_size, noise_level):

        def func_hills(x, split=0, vals=(1, 1, 10), rev=False):
            ans = np.full(len(x), np.nan)
            if not rev:
                ans[x < split] = vals[0] + np.sin(vals[1] * x[x < split])
                eps = (vals[1] / vals[2]) * np.cos(vals[1] * split) / np.cos(vals[2] * split)
                delta = vals[0] + np.sin(vals[1] * split) - eps * np.sin(vals[2] * split)
                ans[x >= split] = delta + eps * np.sin(vals[2] * x[x >= split])
            else:
                ans[x > split] = vals[0] + np.sin(vals[1] * x[x > split])
                eps = (vals[1] / vals[2]) * np.cos(vals[1] * split) / np.cos(vals[2] * split)
                delta = vals[0] + np.sin(vals[1] * split) - eps * np.sin(vals[2] * split)
                ans[x <= split] = delta + eps * np.sin(vals[2] * x[x <= split])
            return ans
        
        noise = np.random.normal(0, noise_level, sample_size)
        design = np.random.uniform(-2.5, 2.5, size=(sample_size, 30))

        f1 = func_hills(design[:,0], 0, (1, 1, 12))
        f2 = func_hills(design[:,1], 1, (1, 2, 8))
        f3 = func_hills(design[:,2], -1, (0, 3, 15), rev=True)
        f4 = func_hills(design[:,3], 1, (0, 2.5, 10), rev=True)
        true_signal = f1 + f2 + f3 + f4

        response = true_signal + noise

        return design, response, true_signal 
    
    @staticmethod
    def additive_linear(sample_size, noise_level):

        def f1_lin(x):
            knots = [-2.5, -2.3, 1, 2.5]  
            values = [0.5, -2.5, 1.8, 2.3]
            return np.interp(x, knots, values)

        def f2_lin(x):
            knots = [-2.5, -2, -1, 1, 2, 2.5]  
            values = [-0.5, 2.5, 1, -0.5, -2.2, -2.3]
            return np.interp(x, knots, values)

        def f3_lin(x):
            knots = [-2.5, -1.5, 0.5, 2.5]  
            values = [0, -3, 2.5, -1]  
            return np.interp(x, knots, values)

        def f4_lin(x):
            knots = [-2.5, -1.8, -0.5, 1.5, 1.8, 2.5]  
            values = [-1, -3.8, -1, -2.3, -0.5, 0.8]
            return np.interp(x, knots, values)


        noise = np.random.normal(0, noise_level, sample_size)
        design = np.random.uniform(-2.5, 2.5, size=(sample_size, 30))

        y1 = f1_lin(design[:,0])
        y2 = f2_lin(design[:,1])
        y3 = f3_lin(design[:,2])
        y4 = f4_lin(design[:,3])
        true_signal = y1 + y2 + y3 + y4
        response = true_signal + noise


        return design, response, true_signal 

    @staticmethod   
    def additive_step(sample_size, noise_level):

        def func_step(X, knots, vals):
            """Apply piecewise constant values based on knots."""
            # Start with the last value for all x (assuming x > last knot)
            y = np.full_like(X, vals[-1], dtype=float)

            # Assign values for intervals defined by knots
            for i in range(len(knots)):
                if i == 0:
                    y[X <= knots[i]] = vals[i]
                else:
                    y[(X > knots[i - 1]) & (X <= knots[i])] = vals[i]

            return y
        
        def f1(X):
            knots = [-2.3, -1.8, -0.5, 1.1]
            vals = [-3, -2.5, -1, 1, 1.8]
            return func_step(X, knots, vals)


        def f2(X):
            knots = [-2, -1, 1, 2]
            vals = [3, 1.4, 0, -1.7, -1.8]
            return func_step(X, knots, vals)


        def f3(X):
            knots = [-1.5, 0.5]
            vals = [-3.3, 2.5, -1]
            return func_step(X, knots, vals)


        def f4(X):
            knots = [-1.7, -0.4, 1.5, 1.9]
            vals = [-2.8, 0.3, -1.4, 0.4, 1.8]
            return func_step(X, knots, vals)
    
        noise = np.random.normal(0, noise_level, sample_size)
        design = np.random.uniform(-2.5, 2.5, size=(sample_size, 30))

            # Apply functions
        y1 = f1(design[:,0])
        y2 = f2(design[:,1])
        y3 = f3(design[:,2])
        y4 = f4(design[:,3])

        true_signal = y1 + y2 + y3 + y4
        response = true_signal + noise        

        return design, response, true_signal 

    @staticmethod
    def diagonal_data(sample_size, type="supersmooth"):
        indices = np.arange(sample_size) + 1
        design = dia_matrix(np.diag(1 / np.sqrt(indices)))
        if type == "supersmooth":
            true_signal = 5 * np.exp(-0.1 * indices)
        elif type == "smooth":
            true_signal = 5000 * np.abs(np.sin(0.01 * indices)) * indices ** (-1.6)
        elif type == "rough":
            true_signal = 250 * np.abs(np.sin(0.002 * indices)) * indices ** (-0.8)
        else:
            ValueError("Currently, only supersmooth, smooth and rough are supported.")

        response_noiseless = design @ true_signal
        return design, response_noiseless, true_signal

    @staticmethod
    def beta_sparse(sample_size, power=3, factor=10):
        true_signal = 1 / (1 + np.arange(sample_size)) ** power
        true_signal = factor * true_signal / np.sum(np.abs(true_signal))

        cov = np.identity(sample_size)

        design = np.random.multivariate_normal(np.zeros(sample_size), cov, sample_size)
        response_noiseless = design @ true_signal

        return design, response_noiseless, true_signal

    @staticmethod
    def s_sparse(sample_size=1000, s=[20, 40, 60], alpha=[1, 0.5, 0.25], factor=10):
        if (s == None) or (alpha == None):
            raise ValueError("s and alpha must be specified.")
        if len(s) != len(alpha):
            raise ValueError("The length of s and alpha must be equal.")
        if (len(s) == 0) or (len(alpha) == 0):
            raise ValueError("s and alpha must not be empty.")

        true_signal = np.zeros(sample_size)

        start = 0
        for i in range(len(s)):
            end = s[i]
            true_signal[start:end] = alpha[i]
            start = end  # Update start for the next segment

        true_signal = factor * true_signal / np.sum(np.abs(true_signal))

        cov = np.identity(sample_size)

        design = np.random.multivariate_normal(np.zeros(sample_size), cov, sample_size)
        response_noiseless = design @ true_signal

        return design, response_noiseless, true_signal

    @staticmethod
    def gravity(sample_size, a=0, b=1, d=0.25):
        """
        Discretisation of a 1-D model problem in gravity surveying, in which
        a mass distribution :math:`f(t)` is located at depth d, while the vertical
        component of the gravity field :math:`g(s)` is measured at the surface.

        The resulting problem is a first-kind Fredholm integral equation  with kernel

        .. math::
            K(s,t) = d(d^2 + (s-t)^2)^{-3/2},

        with the right-hand side given by

        .. math::
            f(t) = \\sin(\\pi t) + 0.5\\sin(2\\pi t).

        The problem is discretized by means of the midpoint quadrature rule.

        **Parameters**

        *sample_size*: ``int``. Specifies the size of the design matrix to be generated.

        *a*: ``int``. The lower bound of the integration interval.

        *b*: ``int``. The upper bound of the integration interval.

        *d*: ``int``. The depth at which the magnetic deposit is located.

        **Returns**

        *design*: ``ndarray``. The design matrix.

        *response_noiseless*: ``ndarray``. The noiseless response. The response is produced by applying the design to the signal.

        *true_signal*: ``ndarray``. The true signal.
        """

        # Parameter controlling the ill-posedness: the larger, the more ill-posed, default in regtools: d = 0.25

        t = (np.arange(1, sample_size + 1) - 0.5) / sample_size
        s = ((np.arange(1, sample_size + 1) - 0.5) * (b - a)) / sample_size
        T, S = np.meshgrid(t, s)

        design = (1 / sample_size) * d * (d**2 * np.ones((sample_size, sample_size)) + (S - T) ** 2) ** (-(3 / 2))
        true_signal = np.sin(np.pi * t) + 0.5 * np.sin(2 * np.pi * t)

        response_noiseless = design @ true_signal
        return design, response_noiseless, true_signal

    @staticmethod
    def heat(sample_size, kappa=1):
        """
        A first kind Volterra integral equation with :math:`[0,1]` as integration interval. The kernel is :math:`K(s,t) = k(s-t)` with the heat kernel

        .. math::
            k(t) = \\frac{t^{-1/2}}{2\\kappa \\sqrt{\\pi}}\\exp\\left(-\\frac{1}{4 \\kappa^2 t}\\right) .

        **Parameters**

        *sample_size*: ``int``. Specifies the size of the design matrix to be generated.

        *kappa*: ``int``. Here, kappa controls the ill-conditioning of the matrix:

        - :math:`\\kappa = 5` gives a well-conditioned problem.

        - :math:`\\kappa = 1` gives an ill-conditioned problem.

        **Returns**

        *design*: ``ndarray``. The design matrix.

        *response_noiseless*: ``ndarray``. The noiseless response. An exact soltuion is constructed, and the response is produced by applying the design to the signal.

        *true_signal*: ``ndarray``. The true signal.
        """

        # Initialization
        h = 1 / sample_size
        t = np.linspace(h / 2, 1, sample_size)  # midpoints
        c = h / (2 * kappa * np.sqrt(np.pi))
        d = 1 / (4 * kappa**2)

        # Compute the matrix A
        k = c * t ** (-1.5) * np.exp(-d / t)
        r = np.zeros(len(t))
        r[0] = k[0]
        design = toeplitz(k, r)

        # Compute the vectors x and b
        true_signal = np.zeros(sample_size)
        for i in range(1, sample_size // 2 + 1):
            ti = i * 20 / sample_size
            if ti < 2:
                true_signal[i - 1] = 0.75 * (ti**2) / 4
            elif ti < 3:
                true_signal[i - 1] = 0.75 + (ti - 2) * (3 - ti)
            else:
                true_signal[i - 1] = 0.75 * np.exp(-(ti - 3) * 2)

        true_signal[sample_size // 2 + 1 :] = 0

        response_noiseless = design @ true_signal

        return design, response_noiseless, true_signal

    @staticmethod
    def deriv2(sample_size, example=1):
        """
        Mildly ill-posed inverse problem based on the discretization of a
        Fredholm integral equation of the first kind whose kernel K is the
        Green's function for the second derivative:

        .. math::
             K(s,t) = \\begin{cases}s(t-1)  ,  &s <  t, \\\\ t(s-1)  ,  &s \\geq t.\\end{cases}

        The right hand side :math:`g` and the solution :math:`f` can be chosen as follows:

        1. :math:`g(s) = (s^3 - s)/6` and :math:`f(t) = t`.

        2. :math:`g(s) = \\exp(s) + (1-e)s - 1` and :math:`f(t) = \\exp(t)`.

        3. :math:`g(s) = \\begin{cases}(4s^3 - 3s)/24,  &s <  0.5,\\\\(-4s^3 + 12s^2 - 9s + 1)/24,  &s \\geq 0.5.\\end{cases}` and :math:`f(t) = \\begin{cases}t,  &t <  0.5,\\\\1-t,  &t \\geq 0.5.\\end{cases}`.


        **Parameters**

        *sample_size*: ``int``. Specifies the size of the design matrix to be generated.

        *example*: ``int``. The example to be used. Must be 1, 2 or 3.

        **Returns**

        *design*: ``ndarray``. The design matrix.

        *response_noiseless*: ``ndarray``. The noiseless response. (Based on discretisation).

        *true_signal*: ``ndarray``. The true signal.
        """
        # Initialize variables and compute coefficients
        h = 1 / sample_size
        sqh = np.sqrt(h)
        h32 = h * sqh
        h2 = h**2
        sqhi = 1 / sqh
        t = 2 / 3
        design = np.zeros((sample_size, sample_size))

        # Compute the matrix A
        for i in range(1, sample_size + 1):
            design[i - 1, i - 1] = h2 * ((i**2 - i + 0.25) * h - (i - t))
            for j in range(1, i):
                design[i - 1, j - 1] = h2 * (j - 0.5) * ((i - 0.5) * h - 1)
        design = design + np.tril(design, -1).T

        # Compute the right-hand side vector b
        response_noiseless = np.zeros(sample_size)
        if example == 1:
            for i in range(1, sample_size + 1):
                response_noiseless[i - 1] = h32 * (i - 0.5) * ((i**2 + (i - 1) ** 2) * h2 / 2 - 1) / 6
        elif example == 2:
            ee = 1 - np.exp(1)
            for i in range(1, sample_size + 1):
                response_noiseless[i - 1] = sqhi * (np.exp(i * h) - np.exp((i - 1) * h) + ee * (i - 0.5) * h2 - h)
        elif example == 3:
            if sample_size % 2 != 0:
                raise ValueError("Order n must be even")
            else:
                for i in range(1, sample_size // 2 + 1):
                    s1 = i * h
                    s12 = s1**2
                    s2 = (i - 1) * h
                    s22 = s2**2
                    response_noiseless[i - 1] = sqhi * (s12 + s22 - 1.5) * (s12 - s22) / 24
                for i in range(sample_size // 2 + 1, sample_size + 1):
                    s1 = i * h
                    s12 = s1**2
                    s2 = (i - 1) * h
                    s22 = s2**2
                    response_noiseless[i - 1] = (
                        sqhi * (-(s12 + s22) * (s12 - s22) + 4 * (s1**3 - s2**3) - 4.5 * (s12 - s22) + h) / 24
                    )
        else:
            raise ValueError("Illegal value of example")

        # Compute the solution vector x
        true_signal = np.zeros(sample_size)
        if example == 1:
            for i in range(1, sample_size + 1):
                true_signal[i - 1] = h32 * (i - 0.5)
        elif example == 2:
            for i in range(1, sample_size + 1):
                true_signal[i - 1] = sqhi * (np.exp(i * h) - np.exp((i - 1) * h))
        elif example == 3:
            for i in range(1, sample_size // 2 + 1):
                true_signal[i - 1] = sqhi * ((i * h) ** 2 - ((i - 1) * h) ** 2) / 2
            for i in range(sample_size // 2 + 1, sample_size + 1):
                true_signal[i - 1] = sqhi * (h - ((i * h) ** 2 - ((i - 1) * h) ** 2) / 2)

        return design, response_noiseless, true_signal

    @staticmethod
    def phillips(sample_size):
        """
        Discretization of the *famous* first-kind Fredholm integral equation deviced by D. L. Phillips.

        Define the function

        .. math::
             \\phi(x) = \\begin{cases}1 + \\cos(x\\pi/3), & |x| < 3,\\\\0, & |x| \\geq 3.\\end{cases}

        The kernel K, the solution f, and the right-hand side g are given by:

        - :math:`K(s,t) = \\phi(s-t)`

        - :math:`f(t) = \\phi(t)`

        - :math:`g(s) = (6-|s|)(1+0.5 \\cos(s\\pi/3)) + 9/(2\\pi)\\sin(|s|\\pi/3)`

        Both integration intervals are [-6,6].

        **Parameters**

        *sample_size*: ``int``. Specifies the size of the design matrix to be generated. Must be a multiple of 4.

        **Returns**

        *design*: ``ndarray``. The design matrix.

        *response_noiseless*: ``ndarray``. The noiseless response. (obtained from discretisation).

        *true_signal*: ``ndarray``. The true signal.
        """
        # Check if n is a multiple of 4
        if sample_size % 4 != 0:
            raise ValueError("The order n must be a multiple of 4")

        # Compute the matrix A using the toeplitz function
        h = 12 / sample_size
        n4 = sample_size // 4
        c = np.cos(np.arange(-1, n4 + 1) * 4 * np.pi / sample_size)
        r1 = np.zeros(sample_size)
        # Debug: Print array shapes

        r1[:n4] = h + 9 / (h * np.pi**2) * (2 * c[1 : n4 + 1] - c[:n4] - c[2 : n4 + 2])
        r1[n4] = h / 2 + 9 / (h * np.pi**2) * (np.cos(4 * np.pi / sample_size) - 1)
        design = toeplitz(r1)

        # Compute the right-hand side b
        response_noiseless = np.zeros(sample_size)
        c = np.pi / 3

        for i in range(sample_size // 2, sample_size):
            t1 = -6 + (i + 1) * h
            t2 = t1 - h
            response_noiseless[i] = (
                t1 * (6 - abs(t1) / 2)
                + ((3 - abs(t1) / 2) * np.sin(c * t1) - 2 / c * (np.cos(c * t1) - 1)) / c
                - t2 * (6 - abs(t2) / 2)
                - ((3 - abs(t2) / 2) * np.sin(c * t2) - 2 / c * (np.cos(c * t2) - 1)) / c
            )
            response_noiseless[sample_size - i - 1] = response_noiseless[i]
        response_noiseless /= np.sqrt(h)

        # Compute the solution x
        true_signal = np.zeros(sample_size)
        # s = np.arange(0, h * 5 + 10 * np.finfo(float).eps, h)
        # original version (problem with array sizes in np.arange)
        # s = np.range(n4)
        # s = np.arange(0, h * (n4+1), h)
        s = np.linspace(0, 3, n4 + 1, endpoint=True)
        true_signal[2 * n4 : 3 * n4] = (h + np.diff(np.sin(s * c)) / c) / np.sqrt(h)

        true_signal[n4 : 2 * n4] = true_signal[3 * n4 - 1 : 2 * n4 - 1 : -1]

        return design, response_noiseless, true_signal


class SimulationParameters:
    """
    `[Source] <https://github.com/ESFIEP/EarlyStopping/edit/main/src/EarlyStopping/simulation_wrapper.py>`_


    A class for managing and validating the parameters for simulation in the `SimulationWrapper` class. This class
    ensures the validity of input parameters required for performing simulation tasks and manages attributes
    such as design matrix, noise level, iteration limits, and parallel processing settings.

    **Parameters**

    *design*: ``ndarray``. The design matrix of the simulation, representing the model matrix.

    *true_signal*: ``ndarray``. The true signal or target vector used for generating simulated response values.

    *true_noise_level*: ``float``. The standard deviation of the normally distributed noise applied to the response.

    *monte_carlo_runs*: ``int``. Defines the number of Monte-Carlo runs to perform in the simulation.

    *noise*: ``ndarray, optional``. Specifies an initial noise matrix; defaults to `None`, in which case noise will
    be generated as needed.

    *response_noiseless*: ``ndarray, optional``. Represents the noiseless response vector, if available. Default is `None`.

    *critical_value*: ``float, optional``. A critical threshold value for the simulation, used in criteria like
    early stopping. Default is `None`.

    *interpolation*: ``bool, default=False``. Specifies whether to use interpolation techniques within the simulation.

    *computation_threshold*: ``float, default=10 ** (-8)``. A small threshold to control numerical computations
    in the simulation procedures.

    *cores*: ``int, default=5``. Specifies the number of processor cores to use for parallel execution.
    """

    def __init__(
        self,
        design,
        true_signal,
        true_noise_level,
        monte_carlo_runs,
        noise=None,
        response_noiseless=None,
        critical_value=None,
        interpolation=False,
        computation_threshold=10 ** (-8),
        cores=5,
    ):

        self.design = design
        self.true_signal = true_signal
        self.true_noise_level = true_noise_level
        self.monte_carlo_runs = monte_carlo_runs
        self.noise = noise
        self.response_noiseless = response_noiseless
        self.critical_value = critical_value
        self.interpolation = interpolation
        self.computation_threshold = computation_threshold
        self.cores = cores

        self.__validate()

    def __validate(self):

        XXt = np.transpose(self.design) @ self.design

        if not isinstance(self.true_signal, np.ndarray):
            raise ValueError("true_signal must be a numpy array.")
        if not (0 <= self.true_noise_level):
            raise ValueError("true_noise_level must be greater than or equal to 0.")

        if scipy.sparse.issparse(XXt):
            warnings.warn(
                "PARAMETER WARNING: The design matrix is sparse. Escaping from __validate().",
                category=UserWarning,
            )
            return None
        else:
            warnings.warn(
                "PARAMETER WARNING: The design matrix is NOT sparse.",
                category=UserWarning,
            )
            rank = np.linalg.matrix_rank(XXt)

        if not (rank == XXt.shape[0]):
            warnings.warn(
                "PARAMETER WARNING: The inverse problem is ill-posed, which is currently not fully supported by landweber.",
                category=UserWarning,
            )


class SimulationWrapper:
    """
    `[Source] <https://github.com/ESFIEP/EarlyStopping/edit/main/src/EarlyStopping/simulation_wrapper.py>`_ A wrapper class for collecting montecarlo simulation data.

    **Parameters**

    *design*: ``ndarray``. The design matrix of the simulation, representing the model matrix.

    *true_signal*: ``ndarray``. The true signal or target vector used for generating simulated response values.

    *true_noise_level*: ``float``. The standard deviation of the normally distributed noise applied to the response.

    *monte_carlo_runs*: ``int``. Defines the number of Monte-Carlo runs to perform in the simulation.

    *noise*: ``ndarray, optional``. Specifies an initial noise matrix; defaults to `None`, in which case noise will
    be generated as needed.

    *response_noiseless*: ``ndarray, optional``. Represents the noiseless response vector, if available. Default is `None`.

    *critical_value*: ``float, optional``. Critical value to be used for early stopping. Default is `None` leading to theoretical values being used.

    *interpolation*: ``bool, default=False``. Specifies whether to use interpolation within the simulation.

    *computation_threshold*: ``float, default=10 ** (-8)``. A small threshold to control numerical computations
    in the simulation procedures.

    **Methods**

    +-----------------------------------------------------------------------------------+---------------------------------------------------------+
    |  run_simulation_truncated_svd(``max_iteration``, ``diagonal``, ``data_set_name``) | Run montecarlo simulation for truncated SVD.            |
    +-----------------------------------------------------------------------------------+---------------------------------------------------------+
    |  run_simulation_landweber(``max_iteration``, ``learning_rate``,``data_set_name``) | Run montecarlo simulation for the Landweber iteration.  |
    +-----------------------------------------------------------------------------------+---------------------------------------------------------+
    |  run_simulation_conjugate_gradients(``max_iteration``, ``data_set_name``)         | Run montecarlo simulation for conjugate gradients.      |
    +-----------------------------------------------------------------------------------+---------------------------------------------------------+
    """

    def __init__(
        self,
        design,
        true_signal=None,
        true_noise_level=None,
        monte_carlo_runs=5,
        noise=None,
        response_noiseless=None,
        critical_value=None,
        interpolation=False,
        computation_threshold=10 ** (-8),
        cores=5,
    ):

        self.design = design
        self.true_signal = true_signal
        self.true_noise_level = true_noise_level
        self.monte_carlo_runs = monte_carlo_runs
        self.noise = noise
        self.critical_value = critical_value
        self.interpolation = interpolation
        self.computation_threshold = computation_threshold
        self.cores = cores

        if response_noiseless is None:
            self.response_noiseless = self.design @ self.true_signal
        else:
            self.response_noiseless = response_noiseless

        self.sample_size = design.shape[0]

    def run_simulation_landweber(self, max_iteration, learning_rate=1, data_set_name=None):
        """
        Runs a simulation for an inverse problem using the Landweber iteration method.
        The function generates a noisy response based on the specified noise level and and performs a Monte-Carlo
        simulation to collect various metrics related to the estimator's performance.

        **Parameters**

        *max_iteration*: ``int``. Specifies the maximum number of iterations to use within the simulation.

        *learning_rate*: ``float`` or ``str``, optional. Specifies the learning rate for the Landweber iteration.
        If set to `"auto"`, the method will search for an optimal learning rate. Default is 1.

        *data_set_name*: ``str``, optional. If specified, the results are saved to a CSV file with this name.

        **Returns**

        *results_df*: ``pd.DataFrame``. DataFrame containing the results of the Monte-Carlo simulation.
        """
        info("Running Monte-Carlo simulation for Landweber.")
        if self.noise is None:
            self.noise = np.random.normal(0, self.true_noise_level, (self.sample_size, self.monte_carlo_runs))
        self.response = self.noise + (self.response_noiseless)[:, None]

        if learning_rate == "auto":
            info("Searching for viable learning rates.")
            self.learning_rate = self.search_learning_rate(max_iteration, search_depth=10)
        else:
            self.learning_rate = learning_rate

        results = Parallel(n_jobs=self.cores)(
            delayed(self.monte_carlo_wrapper_landweber)(m, max_iteration) for m in range(self.monte_carlo_runs)
        )

        column_names = [
            "strong_empirical_risk_es",
            "weak_empirical_risk_es",
            "weak_relative_efficiency",
            "strong_relative_efficiency",
            "strong_bias2",
            "strong_variance",
            "strong_risk",
            "weak_bias2",
            "weak_variance",
            "weak_risk",
            "residuals",
            "discrepancy_stop",
            "balanced_oracle_weak",
            "balanced_oracle_strong",
            # "weak_classical_oracle",
            # "strong_classical_oracle",
            # "weak_empirical_oracle",
            # "strong_empirical_oracle"
        ]

        results_df = pd.DataFrame(results, columns=column_names)

        if data_set_name:
            results_df.to_csv(f"{data_set_name}.csv", index=False)

        return results_df

    def run_simulation_truncated_svd(self, max_iteration, diagonal=False, data_set_name=None):
        """
        Runs a simulation for an inverse problem using truncated Singular Value Decomposition (SVD).
        The function generates a noisy response based on the specified noise level and performs a Monte-Carlo
        simulation to collect various metrics related to the estimator's performance.

        **Parameters**

        *max_iteration*: ``int``. Specifies the maximum number of iterations to use within the simulation.

        *diagonal*: ``bool``, optional. Specifies whether to treat the matrix as diagonal. Default is False.

        *data_set_name*: ``str``, optional. If specified, the results are saved to a CSV file with this name.

        **Returns**

        *results_df*: ``pd.DataFrame``. DataFrame containing the results of the Monte-Carlo simulation.

        The resulting DataFrame can be saved to a CSV file if `data_set_name` is provided.
        """
        info("Running Monte-Carlo simulation for TruncatedSVD.")
        self.diagonal = diagonal
        if self.noise is None:
            self.noise = np.random.normal(0, self.true_noise_level, (self.sample_size, self.monte_carlo_runs))

        self.response = self.noise + (self.response_noiseless)[:, None]

        results = Parallel(n_jobs=self.cores)(
            delayed(self.monte_carlo_wrapper_truncated_svd)(m, max_iteration) for m in range(self.monte_carlo_runs)
        )

        # TODO-BS-2024-11-02: Add AIC stop, classical oracles, etc. as column
        column_names = [
            "strong_bias2",
            "strong_variance",
            "strong_risk",
            "weak_bias2",
            "weak_variance",
            "weak_risk",
            "residuals",
            "discrepancy_stop",
            "weak_balanced_oracle",
            "strong_balanced_oracle",
            "weak_classical_oracle",
            "strong_classical_oracle",
            "weak_relative_efficiency",
            "strong_relative_efficiency",
            "weak_error_at_stopping_time",
            "strong_error_at_stopping_time",
        ]

        results_df = pd.DataFrame(results, columns=column_names)

        if data_set_name:
            results_df.to_csv(f"{data_set_name}.csv", index=False)

        return results_df

    def run_simulation_L2_boost(self, max_iteration, data_set_name=None):
        """
        Runs a simulation for a high dimensional linear model using truncated Singular Value Decomposition (SVD).
        The function generates a noisy response based on the specified noise level and performs a Monte-Carlo
        simulation to collect various metrics related to the estimator's performance.

        **Parameters**

        *max_iteration*: ``int``. Specifies the maximum number of iterations to use within the simulation.

        *data_set_name*: ``str``, optional. If specified, the results are saved to a CSV file with this name.

        **Returns**

        *results_df*: ``pd.DataFrame``. DataFrame containing the results of the Monte-Carlo simulation.

        The resulting DataFrame can be saved to a CSV file if `data_set_name` is provided.
        """
        info("Running Monte-Carlo simulation for L2-boost.")
        if self.noise is None:
            self.noise = np.random.normal(0, self.true_noise_level, (self.sample_size, self.monte_carlo_runs))

        self.response = self.noise + (self.response_noiseless)[:, None]

        results = Parallel(n_jobs=self.cores)(
            delayed(self.monte_carlo_wrapper_L2_boost)(m, max_iteration) for m in range(self.monte_carlo_runs)
        )

        column_names = [
            "bias2",
            "stochastic_error",
            "risk",
            "residuals",
            "noise_estimate",
            "discrepancy_time",
            "residual_ratio_time",
            "balanced_oracle",
            "aic_time",
            "two_step_discrepancy_stop",
            "two_step_residual_ratio_stop",
            "risk_at_discrepancy_time",
            "risk_at_balanced_oracle",
            "risk_at_residual_ratio_time",
            "risk_at_aic_time",
            "relative_efficiency_discrepancy",
            "relative_efficiency_residual_ratio",
            "relative_efficiency_aic",
            "relative_efficiency_two_step_discrepancy_stop",
            "relative_efficiency_two_step_residual_ratio_stop"
        ]

        results_df = pd.DataFrame(results, columns=column_names)

        if data_set_name:
            results_df.to_csv(f"{data_set_name}.csv", index=False)

        return results_df

    def monte_carlo_wrapper_L2_boost(self, m, max_iteration):
        info(f"Monte-Carlo run {m + 1}/{self.monte_carlo_runs}.")
        model_L2_boost = L2_boost(design=self.design, response=self.response[:, m], true_signal=self.true_signal)


        model_L2_boost.iterate(max_iteration)

        bias2 = model_L2_boost.bias2
        stochastic_error = model_L2_boost.stochastic_error
        risk = model_L2_boost.risk
        residuals = model_L2_boost.residuals

        noise_estimate = model_L2_boost.get_noise_estimate()
        print(
            noise_estimate
            - np.mean((model_L2_boost.response - model_L2_boost.design @ model_L2_boost.true_signal) ** 2)
        )
        discrepancy_time = model_L2_boost.get_discrepancy_stop(noise_estimate, max_iteration)
        residual_ratio_time = model_L2_boost.get_residual_ratio_stop(max_iteration)
        balanced_oracle = model_L2_boost.get_balanced_oracle(max_iteration)
        aic_time = model_L2_boost.get_aic_iteration()
        two_step_discrepancy_stop = model_L2_boost.get_aic_iteration(max_iteration=discrepancy_time)
        two_step_residual_ratio_stop = model_L2_boost.get_aic_iteration(max_iteration=residual_ratio_time)

        risk_at_discrepancy_time = risk[discrepancy_time]
        risk_at_balanced_oracle = risk[balanced_oracle]
        risk_at_residual_ratio_time = risk[residual_ratio_time]
        risk_at_aic_time = risk[aic_time]
        risk_at_two_step_discrepancy_stop = risk[two_step_discrepancy_stop]
        risk_at_two_step_residual_ratio_stop = risk[two_step_residual_ratio_stop]

        relative_efficiency_discrepancy = np.sqrt(np.min(risk) / risk_at_discrepancy_time)
        relative_efficiency_residual_ratio = np.sqrt(np.min(risk) / risk_at_residual_ratio_time)
        relative_efficiency_aic = np.sqrt(np.min(risk) / risk_at_aic_time)
        relative_efficiency_two_step_discrepancy_stop = np.sqrt(np.min(risk) / risk_at_two_step_discrepancy_stop)
        relative_efficiency_two_step_residual_ratio_stop = np.sqrt(np.min(risk) / risk_at_two_step_residual_ratio_stop)

        return (
            bias2,
            stochastic_error,
            risk,
            residuals,
            noise_estimate,
            discrepancy_time,
            residual_ratio_time,
            balanced_oracle,
            aic_time,
            two_step_discrepancy_stop,
            two_step_residual_ratio_stop,
            risk_at_discrepancy_time,
            risk_at_balanced_oracle,
            risk_at_residual_ratio_time,
            risk_at_aic_time,
            relative_efficiency_discrepancy,
            relative_efficiency_residual_ratio,
            relative_efficiency_aic,
            relative_efficiency_two_step_discrepancy_stop,
            relative_efficiency_two_step_residual_ratio_stop
        )

    def monte_carlo_wrapper_truncated_svd(self, m, max_iteration):
        info(f"Monte-Carlo run {m + 1}/{self.monte_carlo_runs}.")
        model_truncated_svd = TruncatedSVD(
            design=self.design,
            response=self.response[:, m],
            true_signal=self.true_signal,
            true_noise_level=self.true_noise_level,
            diagonal=self.diagonal,
        )

        model_truncated_svd.iterate(max_iteration)

        strong_bias2 = model_truncated_svd.strong_bias2
        strong_variance = model_truncated_svd.strong_variance
        strong_risk = model_truncated_svd.strong_risk
        weak_bias2 = model_truncated_svd.weak_bias2
        weak_variance = model_truncated_svd.weak_variance
        weak_risk = model_truncated_svd.weak_risk
        residuals = model_truncated_svd.residuals

        discrepancy_stop = model_truncated_svd.get_discrepancy_stop(
            self.sample_size * (self.true_noise_level**2), max_iteration
        )
        weak_balanced_oracle = model_truncated_svd.get_weak_balanced_oracle(max_iteration)
        strong_balanced_oracle = model_truncated_svd.get_strong_balanced_oracle(max_iteration)

        weak_classical_oracle = np.argmin(weak_risk)
        strong_classical_oracle = np.argmin(strong_risk)

        weak_error_vector_at_stopping_time = model_truncated_svd.design @ (
            model_truncated_svd.get_estimate(discrepancy_stop) - model_truncated_svd.true_signal
        )
        weak_error_at_stopping_time = np.sum(weak_error_vector_at_stopping_time**2)
        weak_relative_efficiency = np.sqrt(np.min(weak_risk) / weak_error_at_stopping_time)

        strong_error_at_stopping_time = np.sum(
            (model_truncated_svd.get_estimate(discrepancy_stop) - model_truncated_svd.true_signal) ** 2
        )
        strong_relative_efficiency = np.sqrt(np.min(strong_risk) / strong_error_at_stopping_time)

        return (
            strong_bias2,
            strong_variance,
            strong_risk,
            weak_bias2,
            weak_variance,
            weak_risk,
            residuals,
            discrepancy_stop,
            weak_balanced_oracle,
            strong_balanced_oracle,
            weak_classical_oracle,
            strong_classical_oracle,
            weak_relative_efficiency,
            strong_relative_efficiency,
            weak_error_at_stopping_time,
            strong_error_at_stopping_time,
        )

    def run_simulation_conjugate_gradients(self, max_iteration, data_set_name=None):
        """
        Runs a simulation for an inverse problem using the Conjugate Gradients method.
        This function generates a noisy response based on the specified noise level and performs a
        Monte-Carlo simulation to collect various metrics related to the estimator's performance.

        **Parameters**

        *max_iteration*: ``int``. Specifies the maximum number of iterations to use within the simulation.

        *data_set_name*: ``str``, optional. If specified, the results are saved to a CSV file with this name.

        **Returns**

        *results*: ``list``. A list of results from the Monte-Carlo simulation.
        """
        info("Running Monte-Carlo simulation for Conjugate Gradients.")
        if self.noise is None:
            self.noise = np.random.normal(0, self.true_noise_level, (self.sample_size, self.monte_carlo_runs))
        self.response = self.noise + (self.response_noiseless)[:, None]

        results = Parallel(n_jobs=self.cores)(
            delayed(self.monte_carlo_wrapper_conjugate_gradients)(m, max_iteration)
            for m in range(self.monte_carlo_runs)
        )

        column_names = [
            "strong_empirical_oracle",
            "weak_empirical_oracle",
            "discrepancy_stop",
            "strong_empirical_oracle_risk",
            "strong_empirical_stopping_index_risk",
            "weak_empirical_oracle_risk",
            "weak_empirical_stopping_index_risk",
            "squared_residual_at_stopping_index",
            "strong_relative_efficiency",
            "weak_relative_efficiency",
            "terminal_iteration",
        ]

        results_df = pd.DataFrame(results, columns=column_names)

        if data_set_name:
            results_df.to_csv(f"{data_set_name}.csv", index=False)

        return results_df

    def search_learning_rate(self, max_iteration, search_depth):
        u, s, vh = svds(self.design, k=1)
        largest_singular_value = s[0]
        initial_guess_learning_rate = 2 / largest_singular_value**2
        self.learning_rate_candidates = np.array([initial_guess_learning_rate / (2**i) for i in range(search_depth)])
        info("Determine learning rate candidates based on search depth.")

        results = Parallel(n_jobs=self.cores)(
            delayed(self.search_learning_rate_wrapper)(m, max_iteration) for m in range(search_depth)
        )

        learning_rate_candidates_evaluation, strong_errors = zip(*results)

        true_false_vector = np.array(np.vstack(learning_rate_candidates_evaluation).flatten())
        accepted_candidates = self.learning_rate_candidates[true_false_vector]
        accepted_errors = np.array(strong_errors)[true_false_vector]
        return accepted_candidates[np.argmin(accepted_errors)]

    def search_learning_rate_wrapper(self, m, max_iteration):
        model_landweber = Landweber(
            design=self.design,
            response=self.response[:, 0],
            true_signal=self.true_signal,
            true_noise_level=self.true_noise_level,
            learning_rate=self.learning_rate_candidates[m],
        )

        model_landweber.iterate(max_iteration)
        stopping_index_landweber = model_landweber.get_discrepancy_stop(
            self.sample_size * (self.true_noise_level**2), max_iteration
        )

        converges = True
        for k in range((max_iteration - 1)):
            if model_landweber.residuals[k] < model_landweber.residuals[(k + 1)]:
                converges = False

        if converges is False:
            info(f"The estimator does not converge for learning rate: {self.learning_rate_candidates[m]}", color="red")

        if stopping_index_landweber == None:
            stopping_index_landweber = max_iteration

        return converges, model_landweber.strong_empirical_risk[stopping_index_landweber]

    def monte_carlo_wrapper_landweber(self, m, max_iteration):
        info(f"Monte-Carlo run {m + 1}/{self.monte_carlo_runs}.")
        model_landweber = Landweber(
            design=self.design,
            response=self.response[:, m],
            true_signal=self.true_signal,
            true_noise_level=self.true_noise_level,
            learning_rate=self.learning_rate,
        )

        model_landweber.iterate(max_iteration)

        landweber_strong_bias2 = model_landweber.strong_bias2
        landweber_strong_variance = model_landweber.strong_variance
        landweber_strong_risk = model_landweber.strong_risk
        landweber_weak_bias2 = model_landweber.weak_bias2
        landweber_weak_variance = model_landweber.weak_variance
        landweber_weak_risk = model_landweber.weak_risk
        landweber_residuals = model_landweber.residuals

        stopping_index_landweber = model_landweber.get_discrepancy_stop(
            self.sample_size * (self.true_noise_level**2), max_iteration
        )

        balanced_oracle_weak = model_landweber.get_weak_balanced_oracle(max_iteration)
        balanced_oracle_strong = model_landweber.get_strong_balanced_oracle(max_iteration)

        if (balanced_oracle_weak is None) or (balanced_oracle_strong is None):
            raise ValueError("Weak or strong balanced oracle is None. Relative efficiency cannot be computed.")

        if stopping_index_landweber is None:
            stopping_index_landweber = max_iteration
            print("Stopping index is None. Setting to max iteration.")

        landweber_strong_empirical_risk_es = model_landweber.strong_empirical_risk[stopping_index_landweber]
        landweber_weak_empirical_risk_es = model_landweber.weak_empirical_risk[stopping_index_landweber]

        weak_classical_oracle = np.argmin(landweber_weak_risk)
        strong_classical_oracle = np.argmin(landweber_strong_risk)
        weak_empirical_oracle = np.argmin(model_landweber.weak_empirical_risk)
        strong_empirical_oracle = np.argmin(model_landweber.strong_empirical_risk)

        landweber_weak_relative_efficiency = np.sqrt(
            np.min(model_landweber.weak_empirical_risk) / landweber_weak_empirical_risk_es
        )
        landweber_strong_relative_efficiency = np.sqrt(
            np.min(model_landweber.strong_empirical_risk) / landweber_strong_empirical_risk_es
        )

        return (
            landweber_strong_empirical_risk_es,
            landweber_weak_empirical_risk_es,
            landweber_weak_relative_efficiency,
            landweber_strong_relative_efficiency,
            landweber_strong_bias2,
            landweber_strong_variance,
            landweber_strong_risk,
            landweber_weak_bias2,
            landweber_weak_variance,
            landweber_weak_risk,
            landweber_residuals,
            stopping_index_landweber,
            balanced_oracle_weak,
            balanced_oracle_strong,
            # weak_classical_oracle,
            # strong_classical_oracle,
            # weak_empirical_oracle,
            # strong_empirical_oracle
        )

    def monte_carlo_wrapper_conjugate_gradients(self, m, max_iteration):
        info(f"Monte-Carlo run {m + 1}/{self.monte_carlo_runs}.")

        model_conjugate_gradients = ConjugateGradients(
            design=self.design,
            response=self.response[:, m],
            initial_value=None,
            true_signal=self.true_signal,
            true_noise_level=self.true_noise_level,
            computation_threshold=self.computation_threshold,
        )

        if self.critical_value is None:
            self.critical_value = self.sample_size * (self.true_noise_level**2)

        strong_empirical_oracle = model_conjugate_gradients.get_strong_empirical_oracle(
            max_iteration=max_iteration, interpolation=self.interpolation
        )
        weak_empirical_oracle = model_conjugate_gradients.get_weak_empirical_oracle(
            max_iteration=max_iteration, interpolation=self.interpolation
        )
        stopping_index = model_conjugate_gradients.get_discrepancy_stop(
            critical_value=self.critical_value, max_iteration=max_iteration, interpolation=self.interpolation
        )
        strong_empirical_oracle_risk = model_conjugate_gradients.get_strong_empirical_risk(strong_empirical_oracle)
        weak_empirical_oracle_risk = model_conjugate_gradients.get_weak_empirical_risk(weak_empirical_oracle)

        strong_empirical_stopping_index_risk = model_conjugate_gradients.get_strong_empirical_risk(stopping_index)
        weak_empirical_stopping_index_risk = model_conjugate_gradients.get_weak_empirical_risk(stopping_index)

        squared_residual_at_stopping_index = model_conjugate_gradients.get_residual(stopping_index)

        strong_relative_efficiency = np.sqrt(strong_empirical_oracle_risk / strong_empirical_stopping_index_risk)
        weak_relative_efficiency = np.sqrt(weak_empirical_oracle_risk / weak_empirical_stopping_index_risk)

        terminal_iteration = model_conjugate_gradients.iteration

        return (
            strong_empirical_oracle,
            weak_empirical_oracle,
            stopping_index,
            strong_empirical_oracle_risk,
            strong_empirical_stopping_index_risk,
            weak_empirical_oracle_risk,
            weak_empirical_stopping_index_risk,
            squared_residual_at_stopping_index,
            strong_relative_efficiency,
            weak_relative_efficiency,
            terminal_iteration,
        )


def info(message, color="green"):
    if color == "green":
        print(f"\033[92m{message}\033[0m")
    if color == "red":
        print(f"\033[31m{message}\033[0m")
